--=========================================================================--
-- upvalues
--=========================================================================--

local _G, Pokedex, LibStub = _G, Pokedex, LibStub
local format, gsub, strfind, strtrim, strupper = format, gsub, strfind, strtrim, strupper
local assert, loadstring, tonumber, tostring, type = assert, loadstring, tonumber, tostring, type
local ipairs, pairs, select, sort, unpack = ipairs, pairs, select, sort, unpack
local max, mrandom = max, math.random

local CreateFont, CreateFrame, GameTooltip, GameTooltipText, GetCVarBool, SetCVar, UIParent = CreateFont, CreateFrame, GameTooltip, GameTooltipText, GetCVarBool, SetCVar, UIParent
local GetRunningMacro, InterfaceOptionsFrame_OpenToCategory, SecureCmdOptionParse = GetRunningMacro, InterfaceOptionsFrame_OpenToCategory, SecureCmdOptionParse
local CallCompanion, DismissCompanion, Dismount, GetCompanionCooldown, GetCompanionInfo, GetNumCompanions, VehicleExit = CallCompanion, DismissCompanion, Dismount, GetCompanionCooldown, GetCompanionInfo, GetNumCompanions, VehicleExit
local GetCurrentTitle, GetNumTitles, GetTitleName, IsTitleKnown, SetCurrentTitle = GetCurrentTitle, GetNumTitles, GetTitleName, IsTitleKnown, SetCurrentTitle
local GetAchievementInfo, GetLatestCompletedAchievements, GetProfessionInfo, GetProfessions, GetTalentInfo = GetAchievementInfo, GetLatestCompletedAchievements, GetProfessionInfo, GetProfessions, GetTalentInfo
local GetCurrentMapContinent, GetMapContinents, GetRealZoneText, GetSubZoneText, GetWintergraspWaitTime, GetZoneText, SetMapToCurrentZone = GetCurrentMapContinent, GetMapContinents, GetRealZoneText, GetSubZoneText, GetWintergraspWaitTime, GetZoneText, SetMapToCurrentZone
local GetNumPartyMembers, GetNumRaidMembers, SendChatMessage = GetNumPartyMembers, GetNumRaidMembers, SendChatMessage
local GetItemCount, GetSpellInfo, IsSpellKnown, IsUsableSpell = GetItemCount, GetSpellInfo, IsSpellKnown, IsUsableSpell
local InCombatLockdown, IsFlyableArea, IsFlying, IsIndoors, IsInInstance, IsMounted, IsOutdoors, IsSwimming = InCombatLockdown, IsFlyableArea, IsFlying, IsIndoors, IsInInstance, IsMounted, IsOutdoors, IsSwimming
local GetShapeshiftForm, GetUnitSpeed, UnitCanAttack, UnitCastingInfo, UnitInVehicle, UnitIsDead, UnitIsDeadOrGhost, UnitOnTaxi = GetShapeshiftForm, GetUnitSpeed, UnitCanAttack, UnitCastingInfo, UnitInVehicle, UnitIsDead, UnitIsDeadOrGhost, UnitOnTaxi

-- , DruidPostClick, DruidPreClick, ShamanPostClick, ShamanPreClick


local L = LibStub("AceLocale-3.0"):GetLocale("Pokedex")
local BZ = LibStub("LibBabble-Zone-3.0"):GetLookupTable()

--=========================================================================--
-- global variables, constants and types
--=========================================================================--

local IS = Pokedex.Globals.Types.InitStates;       -- Initialization States
local DL = Pokedex.Globals.Types.DebugLevels;      -- Debug Levels
local SL = Pokedex.Globals.Types.SkillLevels;      -- minimum skill level to hold that rank of profession
local MS = Pokedex.Globals.Types.MountSpeeds;      -- Mount Speeds
local MF = Pokedex.Globals.Types.MountFamilies;    -- Mount Families
local DC = {};  -- Debug Categories 
local gc = Pokedex.Globals.Constants;
local gv = Pokedex.Globals.Variables;


--=========================================================================--
-- Addon Management functions
--=========================================================================--

-- Called when the addon is loaded
function Pokedex:OnInitialize()
	self.db = Pokedex.LoadSavedSettings()

	LibStub("AceConfig-3.0"):RegisterOptionsTable("Pokedex", Pokedex.GetUIOptions)
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Pokedex", "Pokedex")
	self:RegisterChatCommand("pd", "ChatCommand")
	self:RegisterChatCommand("pokedex", "ChatCommand")
end


function Pokedex:ChatCommand(input)
	if (gv.InitState ~= IS.INITIALIZED and self:Initialize("ChatCommand") == IS.UNINITIALIZED) then
		self:Print(L["ERROR: Pokedex failed to initialize correctly. This is usually caused when WoW has invalidated its cache and hasn't finished rebuilding it. Please try this action again later."]);
		return
	end

	if (GetRunningMacro()) then
		if (input) then
			local orig = input;
			input = SecureCmdOptionParse(input);
			if (DC.MISC >= DL.AV) then self:Print("Pokedex called from macro, arguments were parsed"); end
			if (DC.MISC >= DL.AV) then self:Print("from:", orig); end
			if (DC.MISC >= DL.AV) then self:Print("to:", input); end
		end
		
		-- popping up options UI from within a macro is most likely not what the user really wanted, so just exit
		if (not input or input:trim() == "") then
			return
		end
	end

	if (not input or input:trim() == "") then
		InterfaceOptionsFrame_OpenToCategory(self.optionsFrame);
	else
		LibStub("AceConfigCmd-3.0").HandleCommand(Pokedex, "pd", "Pokedex", input);
	end
end


-- Called when the addon is enabled
function Pokedex:OnEnable()
	DC = self.db.global.rgDebugInfo;

	-- we should be able to migrate saved data formats immediately, we don't have to wait for init
	self:MigrateSavedData();

	-- check that none of our actions are blocked
	self:RegisterEvent("ADDON_ACTION_BLOCKED");

	-- check that none of our actions are forbidden
	self:RegisterEvent("ADDON_ACTION_FORBIDDEN");

	-- init data tables and register events and hooks
	self:Initialize("OnEnable");
end


-- Called when the addon is disabled
function Pokedex:OnDisable()
	self:UnregisterEvent("ADDON_ACTION_BLOCKED");
	self:UnregisterEvent("ADDON_ACTION_FORBIDDEN");

	if (gv.InitState == IS.INITIALIZED) then
		-- form support in macro may have to adjust to spec changes or new spells learned
		if (gc.rgClass.fDruid or gc.rgClass.fShaman or gc.rgClass.fDeathKnight) then
			self:UnregisterEvent("ACTIVE_TALENT_GROUP_CHANGED");
			self:UnregisterEvent("LEARNED_SPELL_IN_TAB");
			self:UnregisterEvent("PLAYER_TALENT_UPDATE");
		end

		self:UnregisterEvent("COMPANION_LEARNED");

		self:UnregisterEvent("KNOWN_TITLES_UPDATE");
		self:UnregisterEvent("NEW_TITLE_EARNED");
		self:UnregisterEvent("OLD_TITLE_LOST");

		-- skills
		self:UnregisterEvent("ACHIEVEMENT_EARNED");
		self:UnregisterEvent("SKILL_LINES_CHANGED");

		-- dismount
		self:UnregisterEvent("CVAR_UPDATE");
		self:UnregisterEvent("PLAYER_REGEN_DISABLED");
		self:UnregisterEvent("PLAYER_REGEN_ENABLED");
		self:UnregisterEvent("PLAYER_TARGET_CHANGED");

		-- companion autosummon
		self:UnregisterEvent("ZONE_CHANGED_NEW_AREA");

		Pokedex:UnhookAll();
	end
end


-- Initialize all the pet data in the addon
function Pokedex:Initialize(strCaller)
	strCaller = (strCaller or "unknown caller");
	
	if (gv.InitState == IS.INITIALIZED) then 
		if (DC.MISC >= DL.EXCEP) then self:Print("attempt to reinitialize by " .. strCaller); end
		return IS.INITIALIZED;
	end

	if (gv.InitState == IS.INITIALIZING) then 
		if (DC.MISC >= DL.EXCEP) then self:Print("attempt to initialize during initialization by " .. strCaller); end
		return IS.INITIALIZING;
	end

	if (DC.MISC >= DL.EXCEP) then self:Print("initialization triggered by " .. strCaller); end

	-- get skills
	self:GetSkills();

	-- if we can't get names for skill, mounts, companions or titles, then 
	-- we know we're probably in the bad cache state and will have to try
	-- initializing again later
	if ( not self:FUpdateMountInfo() or
	     not self:FUpdateCompanionInfo() or
	     not self:FUpdateTitleInfo() ) then
	     return IS.UNINITIALIZED;
	end
	
	self:UpdateDismountSettings();

	self.db.profile.fFirstBoot = false;

	gv.InitState = IS.INITIALIZING;

	-- create secure action button to support forms
	self:CreateMountButton();

	-- form support in macro may have to adjust to spec changes or new spells learned
	if (gc.rgClass.fDruid or gc.rgClass.fShaman or gc.rgClass.fDeathKnight) then
		self:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED");
		self:RegisterEvent("LEARNED_SPELL_IN_TAB");
		self:RegisterEvent("PLAYER_TALENT_UPDATE");
	end

	-- update mount data if riding skill changes
	self:RegisterEvent("ACHIEVEMENT_EARNED");

	-- update data when new mount or companion added	
	self:RegisterEvent("COMPANION_LEARNED");

	-- keep track of current mount or companion
	self:SecureHook(   "CallCompanion",    "CallCompanionHook");
	self:SecureHook("DismissCompanion", "DismissCompanionHook");

	-- update data when new title added	
	self:RegisterEvent("KNOWN_TITLES_UPDATE");
	self:RegisterEvent("NEW_TITLE_EARNED");
	self:RegisterEvent("OLD_TITLE_LOST");

	-- keep track of current title
	self:SecureHook("SetCurrentTitle", "SetCurrentTitleHook");

	-- update dismount settings if the CVAR gets changed
	self:RegisterEvent("CVAR_UPDATE");
	
	-- allow flying dismount if in combat or if you've targeted something attackable
	self:RegisterEvent("PLAYER_REGEN_DISABLED");
	self:RegisterEvent("PLAYER_REGEN_ENABLED");
	self:RegisterEvent("PLAYER_TARGET_CHANGED");

	-- companion autosummon
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA");

	-- update dismount for gathering if profession gained or lost
	-- LEARNED_SPELL_IN_TAB lets us know when spell is added to general tab, but not when we drop the skill
	-- SPELLS_CHANGED seems like it would get called *all* the time, we'd be a small perf drag on the game
	-- SKILL_LINES_CHANGED even though we look at spells now, this may still be the best way to know that we should
	self:RegisterEvent("SKILL_LINES_CHANGED");

	-- hook tooltip to see if we're about to try to gather
	self:SecureHookScript(GameTooltip, "OnShow", "MainTooltipShow");
	self:SecureHookScript(GameTooltip, "OnHide", "MainTooltipHide");

	if (DC.MISC >= DL.BASIC) then self:EchoCounts(); end
	gv.InitState = IS.INITIALIZED;
end

function Pokedex:SafeCall(fn, ...)
	if (gv.InitState ~= IS.INITIALIZED and self:Initialize("SafeCall") == IS.UNINITIALIZED) then
		self:Print(L["ERROR: Pokedex failed to initialize correctly. This is usually caused when WoW has invalidated its cache and hasn't finished rebuilding it. Please try this action again later."]);
	else
		return fn(self, ...);
	end
end

--=========================================================================--
--=========================================================================--
--
-- DEBUG AND INFRASTRUCTURE FUNCTIONS
--
--=========================================================================--
--=========================================================================--

local function TestSpeed(fVerbose)
	local nSpeed = GetUnitSpeed("player");
	local nPercent = nSpeed - 7;
	local strOut = "";

	if (nSpeed == 0) then
		strOut = format((fVerbose and "%i yards per second" or "%i YPS"), nSpeed);
	elseif (nPercent == 0) then
		strOut = format((fVerbose and "%i yards per second, standard run speed" or "%i YPS"), nSpeed);
	elseif (nPercent > 0) then
		nPercent = ((nSpeed - 7) * 100) / 7;
		strOut = format((fVerbose and "%.2f yards per second, speed boosted by %.2f%%" or "%i YPS  %.2f%%"), nSpeed, nPercent);
	else
		nPercent = (nSpeed * 100) / 7;
		strOut = format((fVerbose and "%.2f yards per second, speed reduced to %.2f%% of run speed" or "%i YPS  %.2f%%"), nSpeed, nPercent);
	end
	
	strOut = gsub(strOut, "%.00", "");
	return strOut;
end

--[===[@debug@
local LibQTip = nil
local ldbTestDataObj = nil
local ldbTestFrame = nil
local HeaderFont = nil
local LineFont = nil

local function UpdateSpeed(frame,elapsed)
	ldbTestDataObj.text = TestSpeed(false);
end

local function TestTip_OnEnter(self)
	-- Acquire a tooltip with 3 columns, respectively aligned to left, center and right
	local tooltip = LibQTip:Acquire("PokedexTestTooltip", 5)
	self.tooltip = tooltip 
	
	tooltip:SetHeaderFont(HeaderFont);
	tooltip:SetFont(LineFont);

	tooltip:AddLine(" ");

	tooltip:AddHeader("Continent", "Zone", "Real Zone", "SubZone");

	SetMapToCurrentZone();
	local iContinent = GetCurrentMapContinent();
	local rgstrContinents = { GetMapContinents() };
	local strContinent = format("%i  %s", iContinent, (rgstrContinents[iContinent] or "unknown"));
		
	tooltip:AddLine(strContinent, GetZoneText(), GetRealZoneText(), GetSubZoneText());
	
	tooltip:AddLine(" ");
	tooltip:AddSeparator();
	tooltip:AddLine(" ");

	tooltip:AddHeader("Flyable", "Swimming", "Outdoors", "Indoors");
	tooltip:AddLine(tostring(IsFlyableArea()), tostring(IsSwimming()), tostring(IsOutdoors()), tostring(IsIndoors()));

	tooltip:AddLine(" ");
	tooltip:AddSeparator();
	tooltip:AddLine(" ");

	tooltip:AddHeader("Mount Type", "Summonable", "Rejected", "N//A");

	local rgMSFiltered = Pokedex.FilterMountSpeeds(Pokedex);
	for i = 1, #rgMSFiltered do
		tooltip:AddLine(gc.rgstrMountSpeedDescs[rgMSFiltered[i]], "X");
	end
	for ms = MS.Flyer, MS.Unknown-1 do
		local fSeen = false;
		for i = 1, #rgMSFiltered do
			if (ms == rgMSFiltered[i]) then fSeen = true; break; end
		end

		if (not fSeen) then
			if (gv.rgMountsByType[ms] ~= nil) then
				tooltip:AddLine(gc.rgstrMountSpeedDescs[ms], nil, "X");
			else
				tooltip:AddLine(gc.rgstrMountSpeedDescs[ms], nil, nil, "X");
			end
		end
	end

	tooltip:AddLine(" ");

	-- Use smart anchoring code to anchor the tooltip to our frame
	tooltip:SmartAnchorTo(self)

	-- Show it
	tooltip:Show()
end
 
local function TestTip_OnLeave(self)
	-- Release the tooltip
	LibQTip:Release(self.tooltip)
	self.tooltip = nil
end

function Pokedex:StartTestUI()
	if (not LibQTip) then
		LibQTip = LibStub('LibQTip-1.0');
	end

	if (not ldbTestFrame) then 
		ldbTestFrame = CreateFrame("Frame", "Pokedex_Test");
		ldbTestFrame:SetScript("OnUpdate", UpdateSpeed);
	end

	if (not ldbTestDataObj) then 
		ldbTestDataObj = LibStub("LibDataBroker-1.1"):NewDataObject("Pokedex_Test", {
			type	= "data source",
			icon	= "Interface\\Icons\\inv_pet_nurturedpenguinegg.png",
			label	= "Pokedex",
			text	= "0 YPS",
			OnEnter = TestTip_OnEnter,
			OnLeave = TestTip_OnLeave,
		});
	end
	
	if (not HeaderFont) then
		HeaderFont = CreateFont("PokedexHeaderFont");
		HeaderFont:SetFont(GameTooltipText:GetFont(), 17);
		HeaderFont:SetTextColor( 255/255, 215/255, 0/255 );
	end
	
	if (not LineFont) then
		LineFont = CreateFont("PokedexLineFont");
		LineFont:SetFont(GameTooltipText:GetFont(), 15);
		LineFont:SetTextColor( 255/255, 250/255, 205/255 );
	end
end
--@end-debug@]===]


function Pokedex:EchoTest()
	if (DC.DISMOUNT >= DL.AV) then 
		self:Printf("fManageAutoDismount=%s  fDismountForGathering=%s  fDismountForCombat=%s  fDismountForAttack=%s", 
			tostring(self.db.profile.fManageAutoDismount), tostring(Pokedex.db.profile.fDismountForGathering), 
			tostring(self.db.profile.fDismountForCombat), tostring(self.db.profile.fDismountForAttack));
	end

--[===[@debug@
	-- table validation - verifies spellId and mount name match - helps catch typos in table
	-- will spam errors if on a taxi and we've ran test code just to get speed
	if (not UnitOnTaxi("player")) then
		local fFlyable, fSwimming, fVashjir, fQiraji = self:GetAreaInfo();
		local count = 0;
		local names = {};
		for k,v in pairs(gc.rgMountAttributes) do
			count = count + 1;
			names[count] = v.name;
			local strName = GetSpellInfo(k);
			if (strName ~= v.name) then
				self:Printf("table mismatch for spellID %s, expected: %s  got: %s", tostring(k), tostring(v.name), tostring(strName));
			end

			if (IsOutdoors()) then
				local fCanUse = IsUsableSpell(k);
				local dms = gv.rgDMS[v.speed];

				if (not fCanUse) then
					if ( dms[MS.Runner] or dms[MS.Walker] or
						(dms[MS.Flyer] and fFlyable) or 
						(dms[MS.Swimmer] and fSwimming) or
						(dms[MS.Vashjir] and fVashjir) or
						(dms[MS.Qiraji]  and fQiraji)) then
						self:Printf("%s oversold as %s", v.name, gc.rgstrMountSpeeds[v.speed]);
					end
				else 
					if ( not dms[MS.Runner] and not dms[MS.Walker] and
						not (dms[MS.Flyer] and fFlyable) and
						not (dms[MS.Swimmer] and fSwimming) and
						not (dms[MS.Vashjir] and fVashjir) and
						not (dms[MS.Qiraji]  and fQiraji)) then
						self:Printf("%s undersold as %s", v.name, gc.rgstrMountSpeeds[v.speed]);
					end
				end
			end
		end
		self:Printf("Pokedex mounts table contains %s entries", tostring(count));

	--[==[
		sort(names);
		for k,v in ipairs(names) do
			self:Printf("%s  %s", tostring(k), tostring(v));
		end
	--]==]
	end

	self:StartTestUI();

--@end-debug@]===]

end


function Pokedex:EchoCounts()
	self:Printf("Current counts are %i mounts, %i companions and %i titles", #self.db.profile.rgMountInfo, #self.db.profile.rgCompanionInfo, #self.db.profile.rgTitleInfo);
end

function Pokedex:EchoSpeed()
	self:Print(TestSpeed(true));
end

function Pokedex:EchoZone()
	SetMapToCurrentZone();

	local rgstrContinents = { GetMapContinents() };
	local strSubZone = GetSubZoneText();
	local rgMSFiltered = self:FilterMountSpeeds();

	if (strSubZone == "") then strSubZone = nil; end
	if (#rgMSFiltered == 0) then 
		rgMSFiltered = { L["no mounts available"] };
	else
		for i = 1, #rgMSFiltered do
			rgMSFiltered[i] = gc.rgstrMountSpeedDescsShort[ rgMSFiltered[i] ];
		end
	end

	local strLocation = self:StrFromVarArg(nil, unpack( { rgstrContinents[GetCurrentMapContinent()], GetZoneText(), strSubZone } ));
	local strMountList = self:StrFromVarArg(" ", unpack(rgMSFiltered));
	
	self:Printf("%s.  Mount order: %s.", strLocation, strMountList);
end

function Pokedex:ADDON_ACTION_BLOCKED(_, culprit, action, ...)
	if (culprit ~= "Pokedex") then return end
	self:Printf("ERROR: ADDON_ACTION_BLOCKED calling %s  %s", action, self:StrFromVarArg(nil, ...));
end

function Pokedex:ADDON_ACTION_FORBIDDEN(_, culprit, action, ...)
	if (culprit ~= "Pokedex") then return end
	self:Printf("ERROR: ADDON_ACTION_BLOCKED calling %s  %s", action, self:StrFromVarArg(nil, ...));
end


function Pokedex:SetDebug(info, value)
	local strName, strValue, index = self:GetArgs(value, 2);
	
	if (strName == nil) then
		self:DebugValues();
		return;
	end

	if (index ~= 1e9) then
		self:Print("ERROR: too many parameters passed");
		self:DebugUsage();
		return;
	end

	local iValue = tonumber(strValue)
	if (iValue == nil) then
		self:Print("ERROR: no value given for debug level");
		self:DebugUsage();
		return;
	elseif (iValue < DL.NONE or iValue > DL.MAX) then
		self:Printf("ERROR: level_value must be number between %i and %i", DL.NONE, DL.MAX);
		self:DebugUsage();
		return;
	end

	strName = strupper(strName);
	if (strName == "ALL") then
		for k in pairs(DC) do
			DC[k] = iValue;
		end
	elseif (DC[strName] ~= nil) then
			DC[strName] = iValue;
	else
		self:Print("ERROR: level_name not recognized. Must match an existing value or ALL.");
		self:DebugUsage();
		return;
	end

	self:DebugValues();
end

function Pokedex:DebugValues()
	self:Print("current debug levels:");
	for k,v in pairs(DC) do
		self:Printf("value: %i  name: %s", v, k);
		-- self:Printf("lvalue:%i pvalue:%i name:%s", DL[v.category], v, k);
	end
end

function Pokedex:DebugUsage()
	self:Print("to see current debug levels:");
	self:Print("  /pd debug ");
	self:Print("to set a level:");
	self:Print("  /pd debug level_name level_value");
end


function Pokedex:PrintTable(table, tableName)
	tableName = tableName or tostring(table);
	if (table == nil) then 
		self:Print("'table' is nil");
	elseif (type(table) ~= "table") then
		self:Printf("'table' is actually of type %s", type(table));
	else
		self:Printf("key/value pairs for table %s", tableName);
		for k,v in pairs(table) do
			self:Printf("key(%s): %s   value(%s): %s", type(k), tostring(k), type(v), tostring(v));
		end
	end
end

function Pokedex:StrFromVarArg(strDelim, ...)
	if (select("#", ...) == 0) then return ""; end
	strDelim = strDelim or "  "

	local strOut = tostring(select(1, ...));
	for i=2, select("#", ...), 1 do
		strOut = format("%s%s%s", strOut, strDelim, tostring(select(i, ...)));
	end

	return strOut;	
end

local function NewCounter(iStart)
	local i = 0
	return function()
		i = i + 1
		return i
	end
end

--=========================================================================--
--=========================================================================--
--
-- BUTTON FUNCTIONS
--
--=========================================================================--
--=========================================================================--

function Pokedex:ACTIVE_TALENT_GROUP_CHANGED()
	self:SetMountButtonMacros();
	LibStub("AceConfigRegistry-3.0"):NotifyChange("Pokedex");
end

function Pokedex:LEARNED_SPELL_IN_TAB(event, spellID)
	if (gc.rgidMonitoredSpell[spellID]) then 
		self:SetMountButtonMacros();
		LibStub("AceConfigRegistry-3.0"):NotifyChange("Pokedex");
	end
end

function Pokedex:PLAYER_TALENT_UPDATE()
	self:SetMountButtonMacros();
	LibStub("AceConfigRegistry-3.0"):NotifyChange("Pokedex");
end

function Pokedex:SetMountButtonMacros()
	local btn = gv.btnToggleMount;

	if (gc.rgClass.fDeathKnight and self.db.profile.fDKCastHornWithMount and IsSpellKnown(gc.idSpellHornOfWinter)) then
		self:SetDeathKnightClicks();
	elseif (gc.rgClass.fDruid and self.db.profile.fDruidUseForms) then
		self:SetDruidClicks();
	elseif (gc.rgClass.fShaman) then
		self:SetShamanClicks();
	else
		self:SetMountClicks();
	end
end

function Pokedex:CreateMountButton()
	gv.btnToggleMount = CreateFrame("Button", "PokedexToggleMountButton", UIParent, "SecureActionButtonTemplate");
	if not gv.btnToggleMount then self:Print("btnToggleMount is nil") end

	_G["BINDING_NAME_CLICK PokedexToggleMountButton:LeftButton"] = L["Toggle Mount"];

	gv.btnToggleMount:SetAttribute("type", "macro");
	self:SetMountButtonMacros();
end

--=========================================================================--
--=========================================================================--
--
-- DEFAULT MOUNT BUTTON FUNCTIONS
--
--=========================================================================--
--=========================================================================--

function Pokedex:SetMountClicks()
	local strMacroWorgen = format([==[
/cast %s
/run Pokedex:SetCurrentMount(%%i);]==], gc.RunningWild.strSpell);

	local strMacroMount = [==[/run Pokedex:SummonPet("MOUNT", %i);]==]

	local strMacroCombat = [==[/Pokedex DismissMount]==]

	local function MountPreClick(btn)
		if InCombatLockdown() then return end

		if (Pokedex:IsMounted()) then
			btn:SetAttribute("macrotext", strMacroCombat);
		else
			local iMount = Pokedex:SelectMount();
			if (iMount == gc.RunningWild.idSpell) then
				btn:SetAttribute("macrotext", format(strMacroWorgen, iMount));
			elseif (iMount > 0) then
				btn:SetAttribute("macrotext", format(strMacroMount, iMount));
			end
		end
	end

	local function MountPostClick(btn)
		if InCombatLockdown() then return end
		btn:SetAttribute("macrotext", strMacroCombat);
	end

	local btn = gv.btnToggleMount;
	btn:SetScript("PreClick",  MountPreClick);
	btn:SetScript("PostClick", MountPostClick);
	MountPostClick(btn);
end

--=========================================================================--
--=========================================================================--
--
-- DEATH KNIGHT MOUNT BUTTON FUNCTIONS
--
--=========================================================================--
--=========================================================================--

function Pokedex:SetDeathKnightClicks()
	-- NOTE: this function assumes that self.db.profile.fDKCastHornWithMount 
	-- and ability to cast Horn of Winter were already checked to be true
	local strSpellHornOfWinter = GetSpellInfo(gc.idSpellHornOfWinter);
	
	local strMacroWorgen = format([==[
/cast [nomounted] %s
/cast %s
/run Pokedex:SetCurrentMount(%%i);]==], strSpellHornOfWinter, gc.RunningWild.strSpell);

	local strMacroMount = format([==[
/cast [nomounted] %s
/run Pokedex:SummonPet("MOUNT", %%i);]==], strSpellHornOfWinter);

	local strMacroCombat = format([==[
/cast [nomounted] %s
/Pokedex DismissMount]==], strSpellHornOfWinter);


	local function DKPreClick(btn)
		if InCombatLockdown() then return end

		if (Pokedex:IsMounted()) then
			btn:SetAttribute("macrotext", strMacroCombat);
		else
			local iMount = Pokedex:SelectMount();
			if (iMount == gc.RunningWild.idSpell) then
				btn:SetAttribute("macrotext", format(strMacroWorgen, iMount));
			elseif (iMount > 0) then
				btn:SetAttribute("macrotext", format(strMacroMount, iMount));
			end
		end
	end

	local function DKPostClick(btn)
		if InCombatLockdown() then return end
		btn:SetAttribute("macrotext", strMacroCombat);
	end

	local btn = gv.btnToggleMount;
	btn:SetScript("PreClick",  DKPreClick);
	btn:SetScript("PostClick", DKPostClick);
	DKPostClick(btn);
end

--=========================================================================--
--=========================================================================--
--
-- SHAMAN MOUNT BUTTON FUNCTIONS
--
--=========================================================================--
--=========================================================================--

local fHasShamanGroundForm = false;

local function FHasInstantGhostWolf()
	local _, _, _, _, _, rank = GetTalentInfo(2, 6);
	return (rank == 2);
end

function Pokedex:SetShamanClicks()

	fHasShamanGroundForm = IsSpellKnown(gc.idSpellGhostWolf);

	local fWaterWalk = self.db.profile.fShamanCastWaterWalkingWithMount and IsSpellKnown(gc.idSpellWaterWalking);
	local fGhostWolf = self.db.profile.fShamanUseGhostWolf and fHasShamanGroundForm;
	local fInstant   = self.db.profile.fShamanUseInstantWolfIfMoving and fGhostWolf and FHasInstantGhostWolf();

	local strWaterWalking = GetSpellInfo(gc.idSpellWaterWalking);
	local strGhostWolf    = GetSpellInfo(gc.idSpellGhostWolf);

	-- fWaterWalk = false;
	-- fGhostWolf = false;
	-- fInstant   = false;

	--=====================================================================--
	-- build macros
	--=====================================================================--

	local strMacroCombat = format("%s%s", 
		fGhostWolf and format("/cast [nomounted] %s\n", strGhostWolf) or "",
		"/Pokedex DismissMount");

	local strMacroSummonMount = format("%s%s%s",
		fGhostWolf and "/cancelform\n" or "",
		fWaterWalk and format("/cast [nomounted, @player] %s\n", strWaterWalking) or "",
		[==[/run Pokedex:SummonPet("MOUNT", %i);]==]);

	--=====================================================================--
	-- logic chunks which will be used in constructing the PreClick handler
	--=====================================================================--

	local strSpeedCheck = fInstant and [==[ or 
      GetUnitSpeed("player") > 0]==] or "";

	local strIndoorCheck = fGhostWolf and [==[ or 
      not IsOutdoors()]==] or "";

	local strNoMountsFallback = fGhostWolf and [==[
else
      strMacro = strGhostWolf
    ]==] or "";

	--=====================================================================--
	-- define PreClick function structure, plug in macros and logic chunks
	--=====================================================================--

	local strPreClickFormat = [===[
function ShamanPreClick(btn)
  if InCombatLockdown() then return end

  local strMacro = ""

  local strGhostWolf = [=[%s]=]

  local strSummonMount = [=[%s]=]

  if (Pokedex:IsMounted()%s%s) then
    strMacro = strGhostWolf
  else
    local iMount = Pokedex:SelectMount();
    if (iMount > 0) then
      strMacro = format(strSummonMount, iMount)
    %send
  end

  -- print("button macro set to \n" .. strMacro)
  if (strMacro == "") then strMacro = "/Pokedex ToggleMount" end

  btn:SetAttribute("macrotext", strMacro)
end
]===]

	local strPreClickFunc = format(strPreClickFormat, strMacroCombat, strMacroSummonMount, strSpeedCheck, strIndoorCheck, strNoMountsFallback)

	--=====================================================================--
	-- define PostClick function structure, plug in combat macro
	--=====================================================================--

	local strPostClickFormat = [===[
function ShamanPostClick(btn)
  if InCombatLockdown() then return end

  local strMacroCombat = [=[
%s]=]

  btn:SetAttribute("macrotext", strMacroCombat)
end
]===]

	local strPostClickFunc = format(strPostClickFormat, strMacroCombat)

	--=====================================================================--
	-- turn strings into functions and set as script handlers
	--=====================================================================--

--[===[@debug@
-- print(format("%s\n%s", "strPreClickFunc",  strPreClickFunc))
-- print(format("%s\n%s", "strPostClickFunc", strPostClickFunc))
--@end-debug@]===]

	assert(loadstring(strPreClickFunc))()
	assert(loadstring(strPostClickFunc))()

	local btn = gv.btnToggleMount;
	btn:SetScript("PreClick",  ShamanPreClick);
	btn:SetScript("PostClick", ShamanPostClick);
	ShamanPostClick(btn);
end


--=========================================================================--
--=========================================================================--
--
-- DRUID MOUNT BUTTON FUNCTIONS
--
--=========================================================================--
--=========================================================================--

local fHasDruidGroundForm = false;

local function FHasFasterCatForm()
	local _, _, _, _, rank = GetTalentInfo(2, 1);
	return (rank > 0);
end

function Pokedex:SetDruidClicks()
	-- NOTE: this function assumes that self.db.profile.fDruidUseForms was already evaluated as true
	
	-- find out which forms we have, stance numbers vary depending on what forms are known
	local cStances = NewCounter();
	local siBear    = IsSpellKnown(gc.idSpellBearForm)    and cStances() or nil;
	local siAquatic = IsSpellKnown(gc.idSpellAquaticForm) and cStances() or nil;
	local siCat     = IsSpellKnown(gc.idSpellCatForm)     and cStances() or nil;
	local siTravel  = IsSpellKnown(gc.idSpellTravelForm)  and cStances() or nil;
	local siCaster  = (IsSpellKnown(gc.idSpellTreeForm)   or IsSpellKnown(gc.idSpellMoonkinForm))     and cStances() or nil;
	local siFlight  = (IsSpellKnown(gc.idSpellFlightForm) or IsSpellKnown(gc.idSpellSwiftFlightForm)) and cStances() or nil;

	local fFastCat = siCat and FHasFasterCatForm();
	local fWorgen  = gc.RunningWild.fHas;
	local fInstant = self.db.profile.fDruidUseFormsIfMoving;

	-- overrides for testing purposes
	-- siAquatic = nil;
	-- siCat     = nil;
	-- siTravel  = nil;
	-- siFlight  = nil;
	-- fWorgen  = true
	-- fFastCat = false
	-- fInstant = false

	-- in addition to the setting, make sure they have an ground form that is faster than running
	local siGround = siTravel or (fFastCat and siCat);
	fInstant = fInstant and siGround;
	fHasDruidGroundForm = siGround;

	-- print(siBear, siAquatic, siCat, siTravel, siCaster, siFlight, siGround)

	-- spell names for use in macros
	local strAquaticForm  = GetSpellInfo(gc.idSpellAquaticForm);
	local strCatForm      = GetSpellInfo(gc.idSpellCatForm);
	local strTravelForm   = GetSpellInfo(gc.idSpellTravelForm);
	local strGroundForm   = (siGround == siTravel) and strTravelForm or strCatForm;
	local strFlightForm   = IsSpellKnown(gc.idSpellSwiftFlightForm) and GetSpellInfo(gc.idSpellSwiftFlightForm) or GetSpellInfo(gc.idSpellFlightForm);

	-- print(fInstant, fFastCat, fTravel, FHasFasterCatForm())
	
	--=====================================================================--
	-- build big macro for use in combat
	--=====================================================================--

	local strMacroCombatCancel = format("%s%s%s", 
		siAquatic and format("[nostance:%i,swimming]", siAquatic) or "", 
		fFastCat  and format("[nostance:%i,noswimming,indoors]", siCat) or "",
		siGround  and format("[nostance:%i/%i,noswimming,noindoors]", siGround, siFlight or siGround) or "");

	local strMacroCombatCast = format("%s%s%s", 
		siAquatic and format("[swimming]%s; ", strAquaticForm) or "", 
		fFastCat  and format("[indoors]%s; ", strCatForm) or "",
		siGround  and format("[%snomounted] %s", siFlight and format("nostance:%i,", siFlight) or "", strGroundForm) or "");

	local strMacroCombat = format("%s%s%s/Pokedex dm",
		(strMacroCombatCancel ~= "") and format("/cancelform %s\n", strMacroCombatCancel) or "",
		(strMacroCombatCast ~= "") and format("/cast %s\n", strMacroCombatCast) or "",
		siFlight and format("/cancelform [stance:%i]\n", siFlight) or "");

	-- print(format("%s\n%s", "new combat macro", strMacroCombat));

	--=====================================================================--
	-- smaller macros used out of combat
	--=====================================================================--

	local strMacroStance = [==[
/Pokedex DismissMount
/cancelform [nostance:%i]
/cast %s]==];

	local strMacroToggleMount = [==[
/cancelform
/Pokedex ToggleMount]==];

	local strMacroDismissMount = [==[/Pokedex DismissMount]==];

	local strMacroSummonMount = [==[
/cancelform
/run Pokedex:SummonPet("MOUNT", %i);]==];

	local strMacroRunningWild = format([==[
/cancelform
/cast %s
/run Pokedex:SetCurrentMount(%%i);]==], gc.RunningWild.strSpell);

	--=====================================================================--
	-- logic chunks which will be used in constructing the PreClick handler
	-- chunks related to scenarios not supported by characters state, such
	-- such as using cat form when indoors, will be ommitted
	--=====================================================================--

	local strFlyingCheck = siFlight and format([==[
if (fFlyable) then 
    if (not fSwimming) then
      strMacro = format(strStance, "%i", "%s")
    else
      strMacro = strToggleMount
    end
  else]==], siFlight, strFlightForm) or "";

	local strSwimmingCheck = siAquatic and format([==[
if (IsUsableSpell(%i) or fSwimming) then
    strMacro = format(strStance, "%i", "%s")
  else]==], gc.idSpellAquaticForm, siAquatic, strAquaticForm) or "";

	local strIndoorCheck = fFastCat and format([==[
if (not IsOutdoors()) then
    strMacro = format(strStance, "%i", "%s")
  else]==], siCat, strCatForm) or "";

	local strSpeedCheck = fInstant and format([==[
if (GetUnitSpeed("player") > 0) then
      strMacro = format(strStance, "%i", "%s")
  else]==], siGround, strGroundForm) or "";

	local strWorgenCheck = fWorgen and format([==[
if (iMount == %i) then 
      strMacro = format(strRunningWild, iMount)
    else]==], gc.RunningWild.idSpell) or "";

	local strNoMountsFallback = siGround and format([==[
else
      strMacro = format(strStance, "%i", "%s")
    ]==], siGround, strGroundForm) or "";

	--=====================================================================--
	-- define PreClick function structure, plug in macros and logic chunks
	--=====================================================================--

	local strPreClickFormat = [===[
function DruidPreClick(btn)
  if InCombatLockdown() then return end

  local fFlyable, fSwimming = Pokedex.GetAreaInfo(Pokedex);
  local strMacro = ""

  local strStance = [=[%s]=]

  local strToggleMount = [=[%s]=]

  local strDismissMount = [=[%s]=]

  local strSummonMount = [=[%s]=]
%s
  %s%s%sif (Pokedex:IsMounted()) then
    strMacro = strDismissMount
  else%s
    local iMount = Pokedex:SelectMount();
    %sif (iMount > 0) then
      strMacro = format(strSummonMount, iMount)
    %send
  end

  -- print("button macro set to \n" .. strMacro)
  if (strMacro == "") then strMacro = strToggleMount end

  btn:SetAttribute("macrotext", strMacro)
end
]===]

	local strPreClickFunc = format(strPreClickFormat, strMacroStance, strMacroToggleMount, strMacroDismissMount, 
		strMacroSummonMount, fWorgen and format("\n  local strRunningWild = [=[\n%s]=]\n", strMacroRunningWild) or "", 
		strFlyingCheck, strSwimmingCheck, strIndoorCheck, strSpeedCheck, strWorgenCheck, strNoMountsFallback);

	--=====================================================================--
	-- define PostClick function structure, plug in combat macro
	--=====================================================================--

	local strPostClickFormat = [===[
function DruidPostClick(btn)
  if InCombatLockdown() then return end

  local strMacroCombat = [=[
%s]=]

  btn:SetAttribute("macrotext", strMacroCombat)
end
]===]

	local strPostClickFunc = format(strPostClickFormat, strMacroCombat)

	--=====================================================================--
	-- turn strings into functions and set as script handlers
	--=====================================================================--

--[===[@debug@
-- print(format("%s\n%s", "strPreClickFunc",  strPreClickFunc))
-- print(format("%s\n%s", "strPostClickFunc", strPostClickFunc))
--@end-debug@]===]

	assert(loadstring(strPreClickFunc))()
	assert(loadstring(strPostClickFunc))()

	local btn = gv.btnToggleMount;
	btn:SetScript("PreClick",  DruidPreClick);
	btn:SetScript("PostClick", DruidPostClick);
	DruidPostClick(btn);
end


--=========================================================================--
--=========================================================================--
--
-- SKILL FUNCTIONS
--
--=========================================================================--
--=========================================================================--

local function SetProf(iProf)
	local name, icon, rank = GetProfessionInfo(iProf);
--[===[@debug@
	if ((name == "Herbalism" and icon ~= gc.strIconHerbalism) or 
	    (name == "Mining"    and icon ~= gc.strIconMining) or 
	    (name == "Skinning"  and icon ~= gc.strIconSkinning)) then
		Pokedex:Printf("icon for %s skill has changed to %s", name, icon);
	end		
--@end-debug@]===]

	if (icon == gc.strIconMining) then 
		gv.fMiner = true;
	elseif (icon == gc.strIconHerbalism) then 
		gv.fHerbalist = true;
	elseif (icon == gc.strIconSkinning) then 
		gv.fSkinner = true;
	end
end

function Pokedex:GetSkills()
	gv.idLastAchievement = GetLatestCompletedAchievements();

	for k,v in pairs(gc.rgRidingIds) do
		local _, strName, _, fHas = GetAchievementInfo(k);
		strName = strName or L["name not yet available"];
		if (DC.MISC >= DL.AV) then self:Printf("for %s riding skill, %s achievement is %s", gc.rgstrSkillRankName[v], strName, tostring(fHas)); end
		if (fHas) then gv.rsRidingSkill = max(gv.rsRidingSkill, v); end
	end

	-- find out what skills are known
	local iProf_1, iProf_2 = GetProfessions();
	if (iProf_1) then
		gv.strProf_1 = GetProfessionInfo(iProf_1);
		SetProf(iProf_1);
	end
	if (iProf_2) then
		gv.strProf_2 = GetProfessionInfo(iProf_2);
		SetProf(iProf_2);
	end

	if (DC.MISC >= DL.BASIC) then self:Printf("SKILLS  Riding:%s  Mining:%s  Herbalism:%s  Skinning:%s", gc.rgstrSkillRankName[gv.rsRidingSkill], tostring(gv.fMiner), tostring(gv.fHerbalist), tostring(gv.fSkinner)); end
end

function Pokedex:ACHIEVEMENT_EARNED()
	local ids = { GetLatestCompletedAchievements() };

	for i = 1, #ids do
		local id = ids[i]
		if (gv.idLastAchievement == id) then break; end

		if (DC.MISC >= DL.AV) then 
			local _, strName = GetAchievementInfo(id);
			strName = strName or L["name not yet available"];
			self:Printf("new achievement learned: %s", strName); 
		end		

		if (gc.rgRidingIds[id] ~= nil) then 
			if (DC.MISC >= DL.BASIC) then self:Printf("achievement for %s riding skill earned", gc.rgstrSkillRankName[ gc.rgRidingIds[id] ]); end
			if (gc.rgRidingIds[id] > gv.rsRidingSkill) then  -- just in case we have achievement spam from some kind of reset
				gv.rsRidingSkill = gc.rgRidingIds[id];
				self:FUpdateMountInfo();
				LibStub("AceConfigRegistry-3.0"):NotifyChange("Pokedex");
			end
		end
	end
	
	gv.idLastAchievement = ids[1];
end

function Pokedex:SKILL_LINES_CHANGED(...)
	-- if (DC.MISC >= DL.AV) then self:Printf("SKILL_LINES_CHANGED  %s", self:StrFromVarArg(nil, ...)); end

	local iProf_1, iProf_2 = GetProfessions();
	if ((not iProf_1 and gv.strProf_1) or (iProf_1 and (gv.strProf_1 ~= GetProfessionInfo(iProf_1))) or 
	    (not iProf_2 and gv.strProf_2) or (iProf_2 and (gv.strProf_2 ~= GetProfessionInfo(iProf_2)))) then

		local fOldMiner, fOldHerbalist, fOldSkinner = gv.fMiner, gv.fHerbalist, gv.fSkinner;
		gv.fMiner, gv.fHerbalist, gv.fSkinner = false, false, false;
		gv.strProf_1, gv.strProf_2 = nil, nil;
		
		-- find out what skills are known
		local iProf_1, iProf_2 = GetProfessions();
		if (iProf_1) then
			gv.strProf_1 = GetProfessionInfo(iProf_1);
			SetProf(iProf_1);
		end
		if (iProf_2) then
			gv.strProf_2 = GetProfessionInfo(iProf_2);
			SetProf(iProf_2);
		end
		
		if (fOldMiner ~= gv.fMiner) then
			if (DC.MISC >= DL.BASIC) then self:Printf("%s Mining skill", gv.fMiner and "gained" or "lost"); end	
		end
		if (fOldHerbalist ~= gv.fHerbalist) then
			if (DC.MISC >= DL.BASIC) then self:Printf("%s Herbalism skill", gv.fHerbalist and "gained" or "lost"); end
		end
		if (fOldSkinner ~= gv.fSkinner) then
			if (DC.MISC >= DL.BASIC) then self:Printf("%s Skinning skill", gv.fSkinner and "gained" or "lost"); end
		end

		LibStub("AceConfigRegistry-3.0"):NotifyChange("Pokedex");
	end

	if (self.db.profile.fDismountForGathering and not (gv.fMiner or gv.fHerbalist or gv.fSkinner)) then
		if (DC.MISC >= DL.BASIC) then self:Print("we no longer have a gathering skill and so will disable dismount for gathering"); end
		self.db.profile.fDismountForGathering = false;
		LibStub("AceConfigRegistry-3.0"):NotifyChange("Pokedex");
	end
end


--=========================================================================--
--=========================================================================--
--
-- MOUNT AND COMPANION SHARED FUNCTIONS
--
--=========================================================================--
--=========================================================================--

function Pokedex:SuperToggle() 
	self:Print(L["This function has been deprecated; please see README.txt for more infomation about the change to ToggleMount"]);
end

function Pokedex:SummonVendor()
	if (IsOutdoors()) then 
		if (gv.iTravelersMammoth ~= 0) then
			if (gv.iTravelersMammoth ~= self:GetCurrentMount()) then self:SummonPet("MOUNT", gv.iTravelersMammoth); end
		elseif (gv.iArgentSquire ~= 0) then
			if (gv.iArgentSquire ~= self:GetCurrentCompanion()) then self:SummonPet("CRITTER", gv.iArgentSquire); end
		else
			self:Print(L["ERROR: You have no mounts or pets with vendor capabilities"]);
		end
	else
		if (gv.iArgentSquire ~= 0) then
			if (gv.iArgentSquire ~= self:GetCurrentCompanion()) then self:SummonPet("CRITTER", gv.iArgentSquire); end
		elseif (gv.iTravelersMammoth ~= 0) then
			self:Print(L["ERROR: You cannot summon your mammoth in this area"]);
		else
			self:Print(L["ERROR: You have no mounts or pets with vendor capabilities"]);
		end
	end
end

function Pokedex:COMPANION_LEARNED(...)
	if (DC.MOUNTS >= DL.BASIC or DC.PETS >= DL.BASIC) then self:Print("COMPANION_LEARNED  " .. self:StrFromVarArg(nil, ...)); end

	self:FUpdateMountInfo()
	self:FUpdateCompanionInfo()
	LibStub("AceConfigRegistry-3.0"):NotifyChange("Pokedex");
end

function Pokedex:CallCompanionHook(type, index)
	if (type == "CRITTER" ) then
		self:SetCurrentCompanion(index)
	elseif (type == "MOUNT") then
		self:SetCurrentMount(index)
	else
		self:Print(L["ERROR: type error"])
	end
end


function Pokedex:DismissCompanionHook(type)
	if (type == "CRITTER" ) then
		self:SetCurrentCompanion(0)
	elseif (type == "MOUNT") then
		self:SetCurrentMount(0)
	else
		self:Print(L["ERROR: type error"])
	end
end


function Pokedex:SummonPet(strType, iPet, fNext, fOther)
	if (iPet == gc.RunningWild.idSpell) then
		if (DC.MOUNTS >= DL.BASIC) then self:Print("ERROR - tried to summon running wild"); end
		return
	end

	local _, _, idSpell = GetCompanionInfo(strType, iPet);

	if (strType == "CRITTER" ) then
		if (self.db.profile.rgCompanionInfo[iPet].id == idSpell) then
			CallCompanion(strType, iPet);
			self:AnnounceSummon(self.db.profile.rgCompanionInfo[iPet].name);
		else
			if (DC.PETS >= DL.BASIC) then self:Print("spellId mismatch between live and saved data; rebuilding and reselecting"); end
			self:FUpdateCompanionInfo();
			self:SummonCompanion(fNext);
		end
	elseif (strType == "MOUNT") then
		if (self.db.profile.rgMountInfo[iPet].id == idSpell) then
			CallCompanion(strType, iPet);
			--self:AnnounceSummon(self.db.profile.rgMountInfo[iPet].name);

			-- change title on load, unless its heroic maloriak temporary title in which case revel in the glory
			if (self.db.profile.fChangeTitleOnMount and GetCurrentTitle() ~= gc.idTitleMinionSlayer) then 
				self:ChangeTitle(); 
			end
		else
			if (DC.MOUNTS >= DL.BASIC) then self:Print("spellId mismatch between live and saved data; rebuilding and reselecting"); end
			self:FUpdateMountInfo();
			self:SummonMount(fNext, fOther);
		end
	else
		self:Print(L["ERROR: type error"]);
	end
end


function Pokedex:AnnounceSummon(strName)
	if (not self.db.profile.fAnnounce) then return end

	local iChannel = self.db.profile.iChannel

	-- if not in raid, try sending it to party
	if (3 == iChannel and 0 == GetNumRaidMembers()) then
		iChannel = 2
	end
	-- if not in a party, just send it to yourself
	if (2 == iChannel and 0 == GetNumPartyMembers()) then
		iChannel = 1
	end

	local strOut
	if (iChannel == 4) then
		strOut = format(L["lets %s know that they have been chosen."], strName);
	else
		strOut = format(L["%s, I choose you!"], strName);
	end

	if (iChannel == 1) then
		self:Print(strOut);
	else
		SendChatMessage(strOut, gc.rgstrChannels[iChannel]);
	end

end


--=========================================================================--
--=========================================================================--
--
-- MOUNT FUNCTIONS
--
--=========================================================================--
--=========================================================================--

function Pokedex:ToggleMount()
	if (Pokedex:IsMounted()) then
		self:DismissMount()
	else
		self:SummonMount()
	end
end

function Pokedex:DismissMount()
	if UnitInVehicle("player") then
		VehicleExit()
	else
		-- TODO find if this goes through dismiss companion hook for tracking
		Dismount()
	end

	if (self.db.profile.fAutoSummonCompanion and self:GetCurrentCompanion() == 0 and #self.db.profile.rgCompanionInfo > 0) then
		self:SummonCompanion()
	end
end

function Pokedex:SummonMount(fNext, fOther)
	local iMount = self:SelectMount(fNext, fOther, true);
	if (iMount > 0) then
		self:SummonPet("MOUNT", iMount, fNext, fOther);
	end
end

function Pokedex:IsMounted()
	return (IsMounted() or UnitInVehicle("player"));
end


-- returns can fly, is swimming, can summon vashjir seahorse, can summon qiraji scarab
function Pokedex:GetAreaInfo()
	--=====================================================================--
	-- Easiest way to know if we can summon specific type of mount is to just
	-- check the spell for summoning a specific example of that type and to 
	-- see if the spell is castable. IsUsableSpell does not care whether you
	-- actually have access to that mount or not. The IsUsableSpell check for 
	-- a flying mount will fail for druids if they're in flight form and 
	-- Auto-Dismount in Flight has been turned off. However, if we know that
	-- we're currently flying, then it has to be a flyable area, right?
	--=====================================================================--
	local fFlyable  = (1 == IsUsableSpell(gc.idSpellRedDrake) or 1 == IsFlying());
	local fSwimming = IsSwimming();
	local fVashjir  = (1 == IsUsableSpell(gc.idSpellSeahorse));
	local fQiraji   = (1 == IsUsableSpell(gc.idSpellRedScarab));

	--=====================================================================--
	-- Abyssal Seahorse is returning true for IsUsable Spell in zones where
	-- the attempt will fail. Blech.
	--=====================================================================--
	if (fVashjir) then
		SetMapToCurrentZone();
		local zone = GetRealZoneText();
		fVashjir = zone == BZ["Abyssal Depths"] or 
				   zone == BZ["Kelp'thar Forest"] or 
				   zone == BZ["Shimmering Expanse"];
	end

	--=====================================================================--
	-- The IsUsableSpell checks will "fail" if the player is a shaman or a
	-- druid, in a form and have the autoUnshift behavior turned off. We
	-- can work around that using /cancelform, but it still requires us to 
	-- have accurate info about what mount types can be used. Fall back to
	-- the old zone based logic.
	--=====================================================================--
	if ((gc.rgClass.fDruid or gc.rgClass.fShaman) and
		(GetShapeshiftForm() > 0) and not GetCVarBool("autoUnshift")) then
		-- print("USING OLD LOCATION BASED DETECTION");

		-- check locations for special rules about what kinds of mounts we can actually summon
		SetMapToCurrentZone();
		local continent = GetCurrentMapContinent();

		fFlyable = IsFlyableArea() and not fSwimming and not IsInInstance()
		
		if (-1 == continent) then 
			fQiraji = strfind(BZ["Ahn'Qiraj"], GetZoneText()); 
		elseif (4 == continent and not fSwimming) then
			-- if they haven't bought cold weather flying, then they can't fly in northrend
			if (not IsSpellKnown(gc.idSpellColdWeatherFlying)) then
				fFlyable = false;
			-- if Wintergrasp is being contested, you can't fly there
			elseif (GetZoneText() == BZ["Wintergrasp"]) then
				fFlyable = (nil ~= GetWintergraspWaitTime());
--[===[@debug@
				if (fFlyable and IsFlyableArea()) then self:Print("Blizz returned correct true for flight status of WG, work around code may no longer be needed") end
--@end-debug@]===]
			end
		elseif (1 == continent or 2 == continent) then 
			if (2 == continent) then 
				if (fFlyable and (GetSubZoneText() == L["Nespirah"])) then
					fFlyable = false;
				end
			end

			if (fFlyable and not IsSpellKnown(gc.idSpellFlightMastersLicense)) then
				fFlyable = false;
			end
		end
	end
	
	return fFlyable, fSwimming, fVashjir, fQiraji;
end

-- returns list of mount speeds eligible for current location
function Pokedex:FilterMountSpeeds(fOther)
	local rgMSFiltered = {};
	local fFlyable, fSwimming, fVashjir, fQiraji = self:GetAreaInfo();

	local function AddType(ms)
		if (gv.rgMountsByType[ms] ~= nil) then 
			rgMSFiltered[#rgMSFiltered+1] = ms;
		end
	end

	-- If user has asked for fOther, then they didn't like the type of mount we chose for them.
	-- The best thing here is to serve up an ATV so that they are mobile no matter where they are.
	-- Should we not actually have any of those then we'll fall back on flipping the flyable state.
	if (fOther) then
		AddType(MS.Hydroplane);
		fFlyable = not fFlyable; 
	end

	if (fFlyable) then 
		AddType(MS.Flyer) 
	end

	if (fSwimming) then
		if (fVashjir) then
			AddType(MS.Vashjir)
		end
		AddType(MS.Swimmer)
		AddType(MS.Hydroplane)
		AddType(MS.Hydropony)
	else
		if (fQiraji) then
			AddType(MS.Qiraji)
		end
		if (self.db.profile.fFavorGroundMounts) then
			AddType(MS.Hydropony)
		end
		AddType(MS.Runner)
		AddType(MS.Walker)
	end
	
	AddType(MS.Unknown)

	return rgMSFiltered;
end


function Pokedex:SelectMount(fNext, fOther, fInSummons)
	local strCurCast = UnitCastingInfo("player")
	if (strCurCast ~= nil) then
		if (DC.MOUNTS >= DL.EXCEP) then self:Printf("currently casting %s; summon will fail for this reason so ignore this summons attempt", strCurCast); end
		return 0;
	end

	-- We may need to search across multiple lists. For example, if the only 
	-- mount in the set was given a rank of 0. Another case is supporting a 
	-- "style over substance" type feature where 310% and 280% mounts are 
	-- treated as the same category.
	local rgMSFiltered = self:FilterMountSpeeds(fOther);

	if (DC.MOUNTS >= DL.AV) then
		self:Print("mount types elligible for this zone:")
		for i=1, #rgMSFiltered, 1 do
			self:Print("  " .. gc.rgstrMountSpeeds[rgMSFiltered[i]])
		end
	end

	local iCurMount = self:GetCurrentMount();
	local rgiFiltered = {};
	local cTotalRanks = 0;
	local fSkippedCur = false;
	local fSkippedHot = false;

	--=====================================================================--
	-- FilterMounts -- Local function that builds list of mounts elligible
	-- for summoning as well as providing a sum of the ranks of those mounts.
	-- Best done as seperate function to support different ways of calling,
	-- next mount versus random mount, and as local function we can reference 
	-- and modify variables in the scope of the calling function.
	-- NOTE: appends to rgiFiltered, so multiple calls can build bigger list
	--=====================================================================--
	local function FilterMounts(msType, iStart, iEnd, fFirstOnly, fInSummons)
		if (iStart == 0 or iStart > iEnd) then
			if (DC.MOUNTS >= DL.EXCEP) then self:Print("invalid search range"); end
			return;
		end 

		if (DC.MOUNTS >= DL.EXCEP) then self:Printf("building list of eligible %s ...", gc.rgstrMountSpeedDescs[msType]); end
		for iMountOfType = iStart, iEnd, 1 do
			-- translate from index in speed group to all mounts index
			local iMount = gv.rgMountsByType[msType].rgIndices[iMountOfType];
			
			-- if its ranked as 0, then it gets skipped
			if (self.db.profile.rgMountInfo[iMount].rank == 0) then
				if (DC.MOUNTS >= DL.EXCEP) then self:Printf("mount %i %s skipped for having rank of 0", iMount, self.db.profile.rgMountInfo[iMount].name); end
			-- cannot call Running Wild from inside any of the old summoning methods
			elseif (fInSummons and iMount == gc.RunningWild.idSpell) then
				if (DC.MOUNTS >= DL.EXCEP) then self:Printf("mount %i %s skipped for not being summonable from this method", iMount, self.db.profile.rgMountInfo[iMount].name); end
			-- if its the current mount, then it is skipped
			elseif (iMount == iCurMount) then
				if (DC.MOUNTS >= DL.EXCEP) then self:Printf("mount %i %s skipped for being current mount", iMount, self.db.profile.rgMountInfo[iMount].name); end
				fSkippedCur = true;
			-- if looking for first good value, then this is it
			elseif (fFirstOnly) then
				rgiFiltered[#rgiFiltered + 1] = iMount;
				cTotalRanks = cTotalRanks + self.db.profile.rgMountInfo[iMount].rank;
				if (DC.MOUNTS >= DL.BASIC) then self:Printf("mount %i %s is Next eligible mount", iMount, self.db.profile.rgMountInfo[iMount].name); end
				return;
			-- if hot mount and hotness is enabled then don't add to list, it will be handled seperately
			elseif (self.db.profile.fEnableHotness and iMount == gv.iHotMount) then
				if (DC.MOUNTS >= DL.EXCEP) then self:Printf("mount %i %s skipped for being hot mount", iMount, self.db.profile.rgMountInfo[iMount].name); end
				fSkippedHot = true;
			-- else its put into the pool of summonables
			else
				rgiFiltered[#rgiFiltered + 1] = iMount;
				cTotalRanks = cTotalRanks + self.db.profile.rgMountInfo[iMount].rank;
				if (DC.MOUNTS >= DL.AV) then self:Printf("mount %i %s added to list of summonable mounts with a rank of %i. Total rank count is %i", iMount, self.db.profile.rgMountInfo[iMount].name, self.db.profile.rgMountInfo[iMount].rank, cTotalRanks); end
			end
		end

		if (DC.MOUNTS >= DL.BASIC) then self:Printf("list built: %i %s with a total rank of %i", #rgiFiltered, gc.rgstrMountSpeedDescs[msType], cTotalRanks); end
	end
	--=====================================================================--
	-- end of local function FilterMounts
	-- resumption of parent function SummonMount
	--=====================================================================--

	-- Generate filtered list of elligible mounts and total ranks.
	for msFilteredType = 1, #rgMSFiltered, 1 do
		local msType = rgMSFiltered[msFilteredType];

		if (fNext and msType == gv.rgMountMap[iCurMount].speed) then
			-- We can optimize our search in this case by starting immediately after current
			FilterMounts(msType, gv.rgMountMap[iCurMount].index+1, #gv.rgMountsByType[msType].rgIndices, true, fInSummons);
			
			-- if none were found in that portion of the list, then start new search from the front of the list
			if (cTotalRanks == 0) then FilterMounts(msType, 1, gv.rgMountMap[iCurMount].index, true, fInSummons); end
		else
			FilterMounts(msType, 1, #gv.rgMountsByType[msType].rgIndices, fNext, fInSummons);
		end
		
		if (#rgiFiltered ~= 0 or fSkippedHot or fSkippedCur) then 
			break;
		end
	end

	-- if we skipped the hot mount while building a list, nows the time for its heat check
	if (fSkippedHot) then
		local iHeatCheck = mrandom(1,100);
		if (self.db.profile.iHotMountPref >= iHeatCheck) then
			if (DC.MOUNTS >= DL.BASIC) then self:Printf("mount %i %s passed heat check (rolled %i, needed %i or less) and will be summoned", gv.iHotMount, self.db.profile.rgMountInfo[gv.iHotMount].name, iHeatCheck, self.db.profile.iHotMountPref); end
			return gv.iHotMount;
		end

		if (DC.MOUNTS >= DL.BASIC) then self:Printf("mount %i %s skipped for failing heat check (rolled %i, needed %i or less) as hot mount", gv.iHotMount, self.db.profile.rgMountInfo[gv.iHotMount].name, iHeatCheck, self.db.profile.iHotMountPref); end
	end

	-- selection returned 0 mounts to choose from	
	if (#rgiFiltered == 0 or cTotalRanks == 0) then 
		-- both values should be in sync
		if (#rgiFiltered ~= cTotalRanks) then self:Print(L["ERROR: only one of #rgiFiltered and cTotalRanks was zero"]); end

		if (fSkippedHot) then
			if (DC.MOUNTS >= DL.BASIC) then self:Printf("hot mount %i %s failed heat check but is apparently only one summonable", gv.iHotMount, self.db.profile.rgMountInfo[gv.iHotMount].name); end
			return gv.iHotMount;
		end

		if (fSkippedCur) then
			if (DC.MOUNTS >= DL.BASIC) then self:Printf("current mount %i %s is apparently only one summonable; doing nothing", iCurMount, self.db.profile.rgMountInfo[iCurMount].name); end
			return 0; -- only summonable mount is already summoned
		end

		-- don't throw a no mounts error if they have a class form for movement
		if (not (gc.rgClass.fShaman and fHasShamanGroundForm and self.db.profile.fShamanUseGhostWolf) and 
		    not (gc.rgClass.fDruid  and fHasDruidGroundForm  and self.db.profile.fDruidUseForms)) then
			self:Print(L["ERROR: You have no summonable mounts."]);
		end
		return 0;
	end

	
	-- only one mount to choose from
	if (#rgiFiltered == 1) then
		if (DC.MOUNTS >= DL.BASIC) then self:Printf("mount %i %s is apparently only one summonable", rgiFiltered[1], self.db.profile.rgMountInfo[rgiFiltered[1]].name); end
		return rgiFiltered[1];
	end		

	-- multiple mounts
	local cRank = mrandom(1,cTotalRanks);
	if (DC.MOUNTS >= DL.BASIC) then self:Printf("random roll from 1 to %i produced %i", cTotalRanks, cRank); end

	for _, iNew in ipairs(rgiFiltered) do
		cRank = cRank - self.db.profile.rgMountInfo[iNew].rank;
		if (DC.MOUNTS >= DL.AV) then self:Printf("mount %i %s's rank of %i brings total down to %i", iNew, self.db.profile.rgMountInfo[iNew].name, self.db.profile.rgMountInfo[iNew].rank, cRank); end
		if (cRank <= 0) then -- found our slot
			return iNew;
		end
	end

	if (cRank > 0) then self:Print(L["ERROR: selection error"]); end
	return 0;
end

-- Hooking CallCompanion and DismissCompanion ensures we'll usually know what 
-- the current state is, but if the user dismisses the mount by clicking off 
-- the buff or casting a spell, we may not get the update, so always get/set 
-- current pet through these functions.
-- TODO: find if there's a way in lua I could scope this variable better.
-- TODO: persist last companion across sessions so that Next starts where you left off
local gr_iLastMountSeen = 0;

function Pokedex:SetCurrentMount(iMount)
	if (DC.MOUNTS >= DL.BASIC) then self:Printf("Current mount set to: %i %s", iMount, self.db.profile.rgMountInfo[iMount].name); end
	gr_iLastMountSeen = iMount;
	self:SetMountForRanking(gr_iLastMountSeen);
end

function Pokedex:GetCurrentMount()
	if (gr_iLastMountSeen == gc.RunningWild.idSpell) then
		if (not gc.RunningWild.fActive) then
			gr_iLastMountSeen = 0; -- might not have an active mount 
		end 
	elseif (gr_iLastMountSeen ~= 0) then
		local _, _, _, _, fActive = GetCompanionInfo("MOUNT", gr_iLastMountSeen);
		if (not fActive) then 
			gr_iLastMountSeen = 0; -- might not have an active mount 
		end 
	end

	if (gr_iLastMountSeen == 0 and gc.RunningWild.fActive) then
		gr_iLastMountSeen = gc.RunningWild.idSpell;
	end
	
	if (gr_iLastMountSeen == 0) then
		for index = 1, GetNumCompanions("MOUNT"), 1 do
			local _, _, _, _, fActive = GetCompanionInfo("MOUNT", index);
			if (fActive) then 
				gr_iLastMountSeen = index;
				break;
			end
		end
	end

	self:SetMountForRanking(gr_iLastMountSeen);
	return gr_iLastMountSeen;
end


function Pokedex:SetMountForRanking(iMount) 
	if (iMount ~= 0) then
		gv.iMountType   = gv.rgMountMap[iMount].speed;
		gv.iMountOfType = gv.rgMountMap[iMount].index;
	elseif (#self.db.profile.rgMountInfo == 0) then
		gv.iMountType   = 0;
		gv.iMountOfType = 0;
	else -- pick first mount of first non-empty category
		for i = 1, MS.Unknown, 1 do
			if (gv.rgMountsByType[i] ~= nil and #gv.rgMountsByType[i]) then
				gv.iMountType   = i;
				gv.iMountOfType = 1;
				break;
			end
		end
	end
end


function Pokedex:FUpdateMountInfo()
	Pokedex:DecomposeMountSpeeds();

	gv.rgstrMountTypes = {};
	gv.iMountType = 0;
	gv.rgMountsByType = {};
	gv.iMountOfType = 0;
	gv.rgstrHotMountNames =  { [0] = L["no hot mount"] };
	gv.iHotMount = 0;
	gv.rgMountMap = { [0] = { speed = 0, index = 0 } };

	-- assumption is that order doesn't change except for insertions of new mounts
	-- so we can speed up searching saved data for rankings by starting the search 
	-- at a predicted index
	local iMatch = 1;
	local iNewMount = 0;
	local iActive = 0;
	local rgNew = { [0] = { name = "NONE", id = 0, rank = gc.iDefaultRank } };

	-- add mount to type categories for which it has a movement mode
	local function AddMountByTypes(dms, index, strMountName)
		for ms,_ in pairs(dms) do
			if (gv.rgMountsByType[ms] == nil) then gv.rgMountsByType[ms] = { rgIndices = {}, rgNames = {} }; end
			local iNew = 1 + #gv.rgMountsByType[ms].rgIndices;

			gv.rgMountsByType[ms].rgIndices[iNew] = index;
			gv.rgMountsByType[ms].rgNames[iNew] = strMountName;

			gv.rgMountMap[index] = { ["speed"] = ms, ["index"] = iNew };
		end
	end
	
	for index = 1, GetNumCompanions("MOUNT"), 1 do
		local _, _, idSpell, _, fActive = GetCompanionInfo("MOUNT", index);
		local strMountName = GetSpellInfo(idSpell);
		if (strMountName == nil) then
			if (gv.InitState == IS.UNINITIALIZED) then
				if (DC.MISC >= DL.BASIC) then self:Print("unable to initialize, missing mount names"); end
			else
				self:Print(L["ERROR: mount name not available"]);
			end
			return false;
		end

		local fMatchMade, iRank = self:GetMountRankFromSavedData(iMatch, idSpell);
		local dms = gv.rgDMS[self:GetMountType(idSpell)];

		-- if (DC.MOUNTS >= DL.AV) then self:Printf("%s  rank: %i  flying: %s  running: %s  special: %s", strMountName, iRank, gc.rgstrMountSpeeds[dms.flying], gc.rgstrMountSpeeds[dms.running], gc.rgstrMountSpeeds[dms.special]); end

		if (fMatchMade) then 
			iMatch = iMatch + 1;
		elseif (not self.db.profile.fFirstBoot) then
			self:Printf(L["New mount added: %s"], strMountName);
			iNewMount = index;
			if (self.db.profile.fEnableHotness and self.db.profile.fNewHotness) then
				self.db.profile.idHotMount = idSpell;
			end
		end

		if (fActive) then iActive = index; end
		if (idSpell == self.db.profile.idHotMount) then gv.iHotMount = index; end
		if (idSpell == gc.idSpellTravelersTundraMammothAlliance or idSpell == gc.idSpellTravelersTundraMammothHorde) then 
			gv.iTravelersMammoth = index;
		end

		rgNew[index] = { name = strMountName, id = idSpell, rank = iRank };
		gv.rgstrHotMountNames[index] = strMountName;
		AddMountByTypes(dms, index, strMountName);
	end

	-- worgen form support
	if (gc.RunningWild.fHas) then
		local iRunningWild = gc.RunningWild.idSpell;
		local strMountName = gc.RunningWild.strSpell;

		local tblTemp = self.db.profile.rgMountInfo[iRunningWild];
		local iRank = tblTemp and tblTemp.rank or gc.iDefaultRank;
		if (not tblTemp and not self.db.profile.fFirstBoot) then
			self:Printf(L["New mount added: %s"], strMountName);
			iNewMount = iRunningWild;
			if (self.db.profile.fEnableHotness and self.db.profile.fNewHotness) then
				self.db.profile.idHotMount = iRunningWild;
			end
		end

		if (gc.RunningWild.fActive) then iActive = iRunningWild; end
		if (iRunningWild == self.db.profile.idHotMount) then gv.iHotMount = iRunningWild; end

		rgNew[iRunningWild] = { name = strMountName, id = iRunningWild, rank = iRank};
		gv.rgstrHotMountNames[iRunningWild] = strMountName;
		AddMountByTypes({[MS.Runner] = true}, iRunningWild, strMountName)
	end

	-- now that we've read everything in, blow away old saved ranking with new list
	self.db.profile.rgMountInfo = rgNew;

	-- if they have no mounts, then we need to make some text available for the drop downs
	if (#self.db.profile.rgMountInfo == 0) then
		gv.rgstrMountTypes = { [0] = L["no mounts available"] }
		gv.rgMountsByType = { [0] = { rgIndices = { [0] = 0 }, rgNames = { [0] = L["no mounts available"] } } }
	-- else, create list of mount types based on those found
	else
		for ms = MS.Unknown, 1, -1 do
			if (gv.rgMountsByType[ms] ~= nil) then
				gv.rgstrMountTypes[ms] = gc.rgstrMountSpeedDescs[ms];
				gv.iMountType = ms;
			end
		end
	end

	-- iActive will be either 0 from when we inited, or was set when we found 
	-- an active mount. Calling SetCurrentCompanion ensures our cached value 
	-- for last active mount is correct, as well as correctly tweaking the 
	-- selected drop down items for the no active mount case.
	-- A brand new mount, though, will trump the active mount for selection.
	self:SetCurrentMount(iActive);
	if (iNewMount ~= 0) then self:SetMountForRanking(iNewMount); end

	return true;
end


-- return values - fMatchMade, iRank
function Pokedex:GetMountRankFromSavedData(iStart, iID)
	local cLimit = #self.db.profile.rgMountInfo;

	if (iStart <= cLimit) then
		for index = iStart, cLimit, 1 do
			if (self.db.profile.rgMountInfo[index].id == iID) then
				return true, self.db.profile.rgMountInfo[index].rank;
			end
		end
		
		for index = 1, iStart - 1, 1 do
			if (self.db.profile.rgMountInfo[index].id == iID) then
				return true, self.db.profile.rgMountInfo[index].rank;
			end
		end
	end

	return false, gc.iDefaultRank;
end


function Pokedex:DecomposeMountSpeeds()
	local fCanFly = (gv.rsRidingSkill >= SL.Expert)

	-- self:Print("DecomposeMountSpeeds");
	gv.rgDMS = {};
	gv.rgDMS[MS.None]          = { [MS.Unknown] = true }
	gv.rgDMS[MS.Flyer]         = { [MS.Flyer] = true }
	gv.rgDMS[MS.Runner]        = { [MS.Runner] = true, [MS.Hydropony] = true }
	gv.rgDMS[MS.Walker]        = { [MS.Walker] = true }
	gv.rgDMS[MS.Swimmer]       = { [MS.Swimmer] = true }
	gv.rgDMS[MS.Vashjir]       = { [MS.Vashjir] = true }
	gv.rgDMS[MS.Qiraji]        = { [MS.Qiraji] = true }
	gv.rgDMS[MS.Unknown]       = { [MS.Unknown] = true }
	gv.rgDMS[MS.Sea_Turtle]    = { [MS.Walker] = true, [MS.Swimmer] = true }
	gv.rgDMS[MS.Black_Scarab]  = { [MS.Runner] = true, [MS.Qiraji] = true }
	gv.rgDMS[MS.Carpet]        = { [MS.Flyer] = true, [MS.Runner] = true }
	gv.rgDMS[MS.ATV]           = fCanFly and { [MS.Flyer] = true, [MS.Runner] = true, [MS.Hydroplane] = true } or { [MS.Runner] = true, [MS.Hydropony] = true };
end


function Pokedex:GetMountType(idSpell)
	if (not gc.rgMountAttributes[idSpell]) then
		-- TODO: tooltip scraping no longer viable, find a new backup method 
		-- (perhaps making guesses based on type from the name, i.e. elekk or gryphon)
		return MS.Unknown;
	end

	return gc.rgMountAttributes[idSpell].speed;
end


--=========================================================================--
--=========================================================================--
--
-- COMPANION FUNCTIONS
--
--=========================================================================--
--=========================================================================--

function Pokedex:ZONE_CHANGED_NEW_AREA(...)
	-- self:Printf("ZONE_CHANGED_NEW_AREA EVENT:  %s", self:StrFromVarArg(nil, ...));
	if (self.db.profile.fAutoSummonCompanion and self:GetCurrentCompanion() == 0 and #self.db.profile.rgCompanionInfo > 0) then
		if not (IsMounted() or UnitInVehicle("player") or UnitOnTaxi("player") or UnitIsDeadOrGhost("player")) then
			self:SummonCompanion()
		end
	end
end

function Pokedex:ToggleCompanion()
	if (self:GetCurrentCompanion() == 0) then
		self:SummonCompanion();
	else
		DismissCompanion("CRITTER");
	end
end

function Pokedex:SummonCompanion(fNext)
	local iCurPet = self:GetCurrentCompanion();

	if (iCurPet ~= 0) then
		local _, duration = GetCompanionCooldown("CRITTER", iCurPet)
		if (duration ~= 0) then
			if (DC.PETS >= DL.EXCEP) then self:Print("in cooldown, ignore this summons attempt"); end
			return
		end
	end

	local fHasSnowballs = (GetItemCount(gc.idItemSnowball) > 0);
	local rgiFiltered = {};
	local cTotalRanks = 0;
	local fSkippedCur = false;
	local fSkippedHot = false;

	--=====================================================================--
	-- FilterCompanions -- Local function that builds list of pets elligible
	-- for summoning as well as providing a sum of the ranks of those pets.
	-- Best done as seperate function to support different ways of calling,
	-- next pet versus random pet, and as local function we can reference 
	-- and modify variables in the scope of the calling function.
	-- NOTE: appends to rgiFiltered, so multiple calls can build bigger list
	--=====================================================================--
	local function FilterCompanions(iStart, iEnd, fFirstOnly)
		if (iStart == 0 or iStart > iEnd) then
			if (DC.PETS >= DL.EXCEP) then self:Print("invalid search range"); end
			return;
		end 

		if (DC.PETS >= DL.EXCEP) then self:Print("building list of eligible pets ..."); end
		for iCompanion = iStart, iEnd, 1 do
			-- if its ranked as 0, then it gets skipped
			if (self.db.profile.rgCompanionInfo[iCompanion].rank == 0) then
				if (DC.PETS >= DL.EXCEP) then self:Printf("companion %i %s skipped for having rank of 0", iCompanion, self.db.profile.rgCompanionInfo[iCompanion].name); end
			else
				-- if we don't have snowballs and this pet requires one, then it is skipped
				if (not fHasSnowballs and gc.rgNeedsSnowball[self.db.profile.rgCompanionInfo[iCompanion].id]) then
					if (DC.PETS >= DL.EXCEP) then self:Printf("companion %i %s skipped for lack of snowballs", iCompanion, self.db.profile.rgCompanionInfo[iCompanion].name); end
				else
					-- if its the current companion, then it is skipped
					if (iCompanion == iCurPet) then
						if (DC.PETS >= DL.EXCEP) then self:Printf("companion %i %s skipped for being current companion", iCompanion, self.db.profile.rgCompanionInfo[iCompanion].name); end
						fSkippedCur = true;
					else
						if (fFirstOnly) then
							-- if looking for first good value, then this is it
							rgiFiltered[#rgiFiltered + 1] = iCompanion;
							cTotalRanks = cTotalRanks + self.db.profile.rgCompanionInfo[iCompanion].rank;
							if (DC.PETS >= DL.BASIC) then self:Printf("companion %i %s is Next eligible pet", iCompanion, self.db.profile.rgCompanionInfo[iCompanion].name); end
							return;
						-- if hot pet and hotness is enabled then don't add to list, it will be handled seperately
						elseif (self.db.profile.fEnableHotness and iCompanion == gv.iHotCompanion) then
							if (DC.PETS >= DL.EXCEP) then self:Printf("companion %i %s skipped for being hot companion", iCompanion, self.db.profile.rgCompanionInfo[iCompanion].name); end
							fSkippedHot = true;
						-- else its put into the pool of summonables
						else
							rgiFiltered[#rgiFiltered + 1] = iCompanion;
							cTotalRanks = cTotalRanks + self.db.profile.rgCompanionInfo[iCompanion].rank;
							if (DC.PETS >= DL.AV) then self:Printf("companion %i %s added to list of summonable pets with a rank of %i. Total rank count is %i", iCompanion, self.db.profile.rgCompanionInfo[iCompanion].name, self.db.profile.rgCompanionInfo[iCompanion].rank, cTotalRanks); end
						end
					end
				end
			end
		end

		if (DC.PETS >= DL.BASIC) then self:Printf("list built: %i pets with a total rank of %i", #rgiFiltered, cTotalRanks); end
	end
	--=====================================================================--
	-- end of local function FilterCompanions
	-- resumption of parent function SummonCompanion
	--=====================================================================--

	-- generate filtered list of elligible pets and total ranks for list
	if (fNext and iCurPet ~= 0) then
		-- We can optimize our search in this case by starting immediately after current
		FilterCompanions(iCurPet+1, #self.db.profile.rgCompanionInfo, true);

		-- if none were found in that portion of the list, then start new search from the front of the list
		if (cTotalRanks == 0) then FilterCompanions(1, iCurPet, true); end
	else
		FilterCompanions(1, #self.db.profile.rgCompanionInfo, fNext);
	end


	-- if we skipped the hot pet while building a list, nows the time for its heat check
	if (fSkippedHot) then
		local iHeatCheck = mrandom(1,100);
		if (self.db.profile.iHotCompanionPref >= iHeatCheck) then
			if (DC.PETS >= DL.BASIC) then self:Printf("companion %i %s passed heat check (rolled %i, needed %i or less) and will be summoned", gv.iHotCompanion, self.db.profile.rgCompanionInfo[gv.iHotCompanion].name, iHeatCheck, self.db.profile.iHotCompanionPref); end
			self:SummonPet("CRITTER", gv.iHotCompanion, fNext);
			return;
		end

		if (DC.PETS >= DL.BASIC) then self:Printf("companion %i %s skipped for failing heat check (rolled %i, needed %i or less) as hot pet", gv.iHotCompanion, self.db.profile.rgCompanionInfo[gv.iHotCompanion].name, iHeatCheck, self.db.profile.iHotCompanionPref); end
	end

	-- selection returned 0 pets to choose from	
	if (#rgiFiltered == 0 or cTotalRanks == 0) then 
		-- both values should be in sync
		if (#rgiFiltered ~= cTotalRanks) then self:Print(L["ERROR: only one of #rgiFiltered and cTotalRanks was zero"]); end

		if (fSkippedHot) then
			if (DC.PETS >= DL.BASIC) then self:Printf("hot companion %i %s failed heat check but is apparently only one summonable", gv.iHotCompanion, self.db.profile.rgCompanionInfo[gv.iHotCompanion].name); end
			self:SummonPet("CRITTER", gv.iHotCompanion, fNext);
			return
		end

		if (fSkippedCur) then
			if (DC.PETS >= DL.BASIC) then self:Printf("current companion %i %s is apparently only one summonable; doing nothing", iCurPet, self.db.profile.rgCompanionInfo[iCurPet].name); end
			return; -- only summonable pet is already summoned
		end

		self:Print(L["ERROR: You have no summonable companions."]);
		return;
	end


	-- only one pet to choose from
	if (#rgiFiltered == 1) then
		if (DC.PETS >= DL.BASIC) then self:Printf("companion %i %s is apparently only one summonable", rgiFiltered[1], self.db.profile.rgCompanionInfo[rgiFiltered[1]].name); end
		self:SummonPet("CRITTER", rgiFiltered[1], fNext);
		return;
	end

	-- multiple pets
	local cRank = mrandom(1,cTotalRanks);
	if (DC.PETS >= DL.BASIC) then self:Printf("random roll from 1 to %i produced %i", cTotalRanks, cRank); end

	for _, iNew in ipairs(rgiFiltered) do
		cRank = cRank - self.db.profile.rgCompanionInfo[iNew].rank;
		if (DC.PETS >= DL.AV) then self:Printf("companion %i %s's rank of %i brings total down to %i", iNew, self.db.profile.rgCompanionInfo[iNew].name, self.db.profile.rgCompanionInfo[iNew].rank, cRank); end
		if (cRank <= 0) then -- found our slot
			self:SummonPet("CRITTER", iNew, fNext);
			return;
		end  
	end

	if (cRank > 0) then self:Print(L["ERROR: selection error"]); end
end


-- Hooking CallCompanion and DismissCompanion should ensure that we'll usually
-- know what the current state is.
-- TODO: find out if trying to cast spell by ID can summon dismiss pets bypassing these checks
-- TODO: find if there's a way in lua I could scope this variable better.
-- TODO: persist last companion across sessions so that Next starts where you left off
local gr_iLastCompanionSeen = 0;

function Pokedex:SetCurrentCompanion(iCompanion)
	if (DC.PETS >= DL.BASIC) then self:Printf("Current companion set to: %i %s", iCompanion, self.db.profile.rgCompanionInfo[iCompanion].name); end
	gr_iLastCompanionSeen = iCompanion;
	self:SetCompanionForRanking(gr_iLastCompanionSeen);
end

function Pokedex:GetCurrentCompanion()
	if (gr_iLastCompanionSeen ~= 0) then
		local _, _, _, _, fActive = GetCompanionInfo("CRITTER", gr_iLastCompanionSeen);
		if (not fActive) then 
			gr_iLastCompanionSeen = 0; -- might not have an active companion 
		end 
	end

	if (gr_iLastCompanionSeen == 0) then
		for index = 1, GetNumCompanions("CRITTER"), 1 do
			local _, _, _, _, fActive = GetCompanionInfo("CRITTER", index);
			if (fActive) then 
				gr_iLastCompanionSeen = index;
				break;
			end
		end
	end

	self:SetCompanionForRanking(gr_iLastCompanionSeen);
	return gr_iLastCompanionSeen;
end


function Pokedex:SetCompanionForRanking(iCompanion) 
	if (iCompanion ~= 0) then
		gv.iCompanionForRanking = iCompanion;
	elseif (#gv.rgstrCompanionNames ~= 0) then
		gv.iCompanionForRanking = 1;
	else
		gv.iCompanionForRanking = 0;
	end
end


function Pokedex:FUpdateCompanionInfo()
	gv.rgstrCompanionNames = {};
	gv.iCompanionForRanking = 0;
	gv.rgstrHotCompanionNames = { [0] = L["no hot companion"] };
	gv.iHotCompanion = 0;

	-- assumption is that order doesn't change except for insertions of new pets
	-- so we can speed up searching saved data for rankings by starting the search 
	-- at a predicted index
	local iMatch = 1;
	local iNewPet = 0;
	local rgNew = { [0] = { name = "NONE", id = 0, rank = gc.iDefaultRank } };

	for index = 1, GetNumCompanions("CRITTER"), 1 do
		local _, _, idSpell, _, fActive = GetCompanionInfo("CRITTER", index);
		local strPetName = GetSpellInfo(idSpell);
		if (strPetName == nil) then
			if (gv.InitState == IS.UNINITIALIZED) then
				if (DC.MISC >= DL.BASIC) then self:Print("unable to initialize, missing companion names"); end
			else
				self:Print(L["ERROR: companion name not available"]);
			end
			return false;
		end

		local fMatchMade, iRank = self:GetCompanionRankFromSavedData(iMatch, idSpell);

		if (fMatchMade) then 
			iMatch = iMatch + 1;
		elseif (not self.db.profile.fFirstBoot) then
			self:Printf(L["New companion added: %s"], strPetName);
			iNewPet = index;
			if (self.db.profile.fEnableHotness and self.db.profile.fNewHotness) then
				self.db.profile.idHotCompanion = idSpell;
			end
		end

		if (fActive) then gv.iCompanionForRanking = index; end
		if (idSpell == self.db.profile.idHotCompanion) then gv.iHotCompanion = index; end
		if (idSpell == gc.idSpellArgentSquire or idSpell == gc.idSpellArgentGruntling) then 
			gv.iArgentSquire = index;
		end

		gv.rgstrCompanionNames[index] = strPetName;
		gv.rgstrHotCompanionNames[index] = strPetName;
		rgNew[index] = { name = strPetName, id = idSpell, rank = iRank };
	end

	-- now that we've read everything in, blow away old saved ranking with new list
	self.db.profile.rgCompanionInfo = rgNew;

	-- if they have no pets, then we need to make some text available for the drop down
	if (#gv.rgstrCompanionNames == 0) then
		gv.rgstrCompanionNames = { [0] = L["no companions available"] };
	end

	-- gv.iCompanionForRanking will be either 0 from when we inited, or was set
	-- when we found an active pet. Calling SetCurrentCompanion ensures our 
	-- cached value for last active pet is correct, as well as correctly 
	-- tweaking the selected drop down item for the no active pet case.
	-- A brand new pet, though, will trump the active pet for selection.
	self:SetCurrentCompanion(gv.iCompanionForRanking);
	if (iNewPet ~= 0) then self:SetCompanionForRanking(iNewPet); end

	return true;
end


-- return values - fMatchMade, iRank
function Pokedex:GetCompanionRankFromSavedData(iStart, iID)
	local cLimit = #self.db.profile.rgCompanionInfo;

	if (iStart <= cLimit) then
		for index = iStart, cLimit, 1 do
			if (self.db.profile.rgCompanionInfo[index].id == iID) then
				return true, self.db.profile.rgCompanionInfo[index].rank;
			end
		end
		
		for index = 1, iStart - 1, 1 do
			if (self.db.profile.rgCompanionInfo[index].id == iID) then
				return true, self.db.profile.rgCompanionInfo[index].rank;
			end
		end
	end

	return false, gc.iDefaultRank;
end


--=========================================================================--
--=========================================================================--
--
-- TITLE FUNCTIONS
--
--=========================================================================--
--=========================================================================--

function Pokedex:ChangeTitle()
	local iCurTitle = gv.rgTitleMap[self:GetCurrentTitle()];

	local rgiFiltered = {};
	local cTotalRanks = 0;
	local fSkippedCur = false;
	local fSkippedHot = false;

	-- Builds list of elligible titles for change. For mounts and pets, this
	-- is a local function; but we have no plans for NextTitle option or any
	-- other scenarios where we have to make multiple passes at the data.
	if (DC.TITLES >= DL.EXCEP) then self:Print("building list of eligible titles ..."); end
	for iTitle = 0, #self.db.profile.rgTitleInfo, 1 do
		-- if its ranked as 0, then it gets skipped
		if (self.db.profile.rgTitleInfo[iTitle].rank == 0) then
			if (DC.TITLES >= DL.EXCEP) then self:Printf("title %i %s skipped for having rank of 0", iTitle, self.db.profile.rgTitleInfo[iTitle].name); end
		else
			-- if its the current title, then it is skipped
			if (iTitle == iCurTitle) then
				if (DC.TITLES >= DL.EXCEP) then self:Printf("title %i %s skipped for being current title", iTitle, self.db.profile.rgTitleInfo[iTitle].name); end
				fSkippedCur = true;
			else
				-- if hot title and hotness is enabled then don't add to list, it will be handled seperately
				if (self.db.profile.fEnableHotness and iTitle == gv.iHotTitle and iTitle ~= 0) then
					if (DC.TITLES >= DL.EXCEP) then self:Printf("title %i %s skipped for being hot title", iTitle, self.db.profile.rgTitleInfo[iTitle].name); end
					fSkippedHot = true;
				-- else its put into the pool of summonables
				else
					rgiFiltered[#rgiFiltered + 1] = iTitle;
					cTotalRanks = cTotalRanks + self.db.profile.rgTitleInfo[iTitle].rank;
					if (DC.TITLES >= DL.AV) then self:Printf("title %i %s added to list of possible titles with a rank of %i. Total rank count is %i", iTitle, self.db.profile.rgTitleInfo[iTitle].name, self.db.profile.rgTitleInfo[iTitle].rank, cTotalRanks); end
				end
			end
		end
	end

	if (DC.TITLES >= DL.BASIC) then self:Printf("list built: %i titles with a total rank of %i", #rgiFiltered, cTotalRanks); end


	-- if we skipped the hot title while building a list, nows the time for its heat check
	if (fSkippedHot) then
		local iHeatCheck = mrandom(1,100);
		if (self.db.profile.iHotTitlePref >= iHeatCheck) then
			if (DC.TITLES >= DL.BASIC) then self:Printf("hot title %i %s passed heat check (rolled %i, needed %i or less) and will be selected", gv.iHotTitle, self.db.profile.rgTitleInfo[gv.iHotTitle].name, iHeatCheck, self.db.profile.iHotTitlePref); end
			SetCurrentTitle(self.db.profile.idHotTitle);
			return;
		end

		if (DC.TITLES >= DL.BASIC) then self:Printf("hot title %i %s failed heat check (rolled %i, needed %i or less) and will not be selected", gv.iHotTitle, self.db.profile.rgTitleInfo[gv.iHotTitle].name, iHeatCheck, self.db.profile.iHotTitlePref); end
	end

	-- selection returned 0 titles to choose from	
	if (#rgiFiltered == 0 or cTotalRanks == 0) then 
		-- both values should be in sync
		if (#rgiFiltered ~= cTotalRanks) then self:Print(L["ERROR: only one of #rgiFiltered and cTotalRanks was zero"]); end

		if (fSkippedHot) then
			if (DC.TITLES >= DL.BASIC) then self:Printf("hot title %i %s failed heat check but is only eligible title", gv.iHotTitle, self.db.profile.rgTitleInfo[gv.iHotTitle].name); end
			SetCurrentTitle(self.db.profile.idHotTitle);
			return
		end

		if (fSkippedCur) then
			if (DC.TITLES >= DL.BASIC) then self:Printf("current title %i %s is only one eligible; doing nothing", iCurTitle, self.db.profile.rgTitleInfo[iCurTitle].name); end
			return; -- only summonable title is already summoned
		end

		self:Print(L["ERROR: You don't have any titles."]);
		return;
	end

	
	-- only one title to choose from
	if (#rgiFiltered == 1) then
		if (DC.TITLES >= DL.BASIC) then self:Printf("title %i %s is apparently only one summonable", rgiFiltered[1], self.db.profile.rgTitleInfo[rgiFiltered[1]].name); end
		SetCurrentTitle(self.db.profile.rgTitleInfo[rgiFiltered[1]].id);
		return;
	end

	-- multiple titles
	local cRank = mrandom(1,cTotalRanks);
	if (DC.TITLES >= DL.BASIC) then self:Printf("random roll from 1 to %i produced %i", cTotalRanks, cRank); end

	for _, iNew in ipairs(rgiFiltered) do
		cRank = cRank - self.db.profile.rgTitleInfo[iNew].rank;
		if (DC.TITLES >= DL.AV) then self:Printf("title %i %s's rank of %i brings total down to %i", iNew, self.db.profile.rgTitleInfo[iNew].name, self.db.profile.rgTitleInfo[iNew].rank, cRank); end
		if (cRank <= 0) then -- found our slot
			SetCurrentTitle(self.db.profile.rgTitleInfo[iNew].id);
			return;
		end
	end

	if (cRank > 0) then self:Print(L["ERROR: selection error"]); end
end


function Pokedex:SetCurrentTitleHook(idTitle)
	if (idTitle < 1) then
		if (DC.TITLES >= DL.BASIC) then self:Printf("No title currently showing"); end
	else
		if (DC.TITLES >= DL.BASIC) then 
			local strTitle = GetTitleName(idTitle);
			self:Printf("Current title set to: %i %s", idTitle, strtrim(strTitle)); 
		end
	end
	gv.iTitleForRanking = gv.rgTitleMap[idTitle];
end

-- return first the title ID and then the sorted order index
function Pokedex:GetCurrentTitle()
	local idTitle = GetCurrentTitle();
	gv.iTitleForRanking = gv.rgTitleMap[idTitle];

	return idTitle, gv.iTitleForRanking;
end

function Pokedex:KNOWN_TITLES_UPDATE()
	if (DC.TITLES >= DL.BASIC) then self:Print("KNOWN_TITLES_UPDATE"); end
	self:FUpdateTitleInfo();
	LibStub("AceConfigRegistry-3.0"):NotifyChange("Pokedex");
end

function Pokedex:NEW_TITLE_EARNED(...)
	if (DC.TITLES >= DL.BASIC) then self:Printf("NTE  %s", self:StrFromVarArg(nil, ...)); end
	self:FUpdateTitleInfo();
	LibStub("AceConfigRegistry-3.0"):NotifyChange("Pokedex");
end

function Pokedex:OLD_TITLE_LOST(...)
	if (DC.TITLES >= DL.BASIC) then self:Printf("OTL  %s", self:StrFromVarArg(nil, ...)); end
	self:FUpdateTitleInfo();
	LibStub("AceConfigRegistry-3.0"):NotifyChange("Pokedex");
end

function Pokedex:FUpdateTitleInfo()
	gv.rgstrTitleNames = {};
	gv.iTitleForRanking = 0;
	gv.rgstrHotTitleNames = { [0] = L["no hot title"] };
	gv.iHotTitle = 0;
	gv.rgTitleMap = {};

	local rgNew = {};
	local iNewTitle = 0;

	local cTitlesKnown = 0;
	for idTitle = 1, GetNumTitles(), 1 do
		if (1 == IsTitleKnown(idTitle)) then
			cTitlesKnown = cTitlesKnown + 1;

			local strTitleName = GetTitleName(idTitle);
			if (strTitleName == nil) then
				if (gv.InitState == IS.UNINITIALIZED) then
					if (DC.MISC >= DL.BASIC) then self:Print("unable to initialize, missing title names"); end
				else
					self:Print(L["ERROR: title name not available"]);
				end
				return false;
			end

			rgNew[cTitlesKnown] = { name = strtrim(strTitleName), id = idTitle, rank = gc.iDefaultRank };
		end
	end
	
	-- if there are no titles known
	if (#rgNew == 0) then
		-- dummy entry holding default rank in empty case
		rgNew[0] = { name = "NONE", id = 0, rank = gc.iDefaultRank };
		gv.rgstrTitleNames[0] = L["no titles available"];

	-- otherwise sort the titles, get saved rank values and get index for hot and current titles
	else
		-- sort this table by name to get a nice drop down order
		sort(rgNew, function(a,b) return (a.name < b.name) end);
		if (DC.TITLES >= DL.AV) then self:Print("Updating Titles ..."); end

		-- Now that we're sorted, lets get our saved rank values. Assumption is that 
		-- order doesn't change except for insertions of new titles so we can speed up 
		-- searching saved data for rankings by starting the search at a predicted index
		-- We'll also take advantage of this iteration through to set our drop down lists.
		local iMatch = 1;
		for index = 1, #rgNew, 1 do
			local fMatchMade, iRank = self:GetTitleRankFromSavedData(iMatch, rgNew[index].id);

			-- hack to work around matron/patron swaps, if match 
			-- didn't happen then try again with the other title ID
			if (not fMatchMade) then
				if (rgNew[index].id == gc.idTitleMatron) then
					fMatchMade, iRank = self:GetTitleRankFromSavedData(iMatch, gc.idTitlePatron);
				elseif (rgNew[index].id == gc.idTitlePatron) then
					fMatchMade, iRank = self:GetTitleRankFromSavedData(iMatch, gc.idTitleMatron);
				end
			end

			if (fMatchMade) then
				rgNew[index].rank = iRank;
				iMatch = iMatch + 1;
			elseif (not self.db.profile.fFirstBoot) then
				self:Printf(L["New title added: %s"], rgNew[index].name)
				gv.iTitleForRanking = index;
				if (self.db.profile.fEnableHotness and self.db.profile.fNewHotness) then
					self.db.profile.idHotTitle = rgNew[index].id;
				end
			end

			if (DC.TITLES >= DL.AV) then self:Printf("index: %i  name: %s  id: %i  rank %i", index, rgNew[index].name, rgNew[index].id, rgNew[index].rank); end
			
			-- save index of hot pet
			if (self.db.profile.idHotTitle == rgNew[index].id) then 
				gv.iHotTitle = index; 
			end

			gv.rgstrTitleNames[index] = rgNew[index].name;
			gv.rgstrHotTitleNames[index] = rgNew[index].name;
		end

		local tblTemp = self.db.profile.rgMountInfo[-1];
		local iRank = tblTemp and tblTemp.rank or 0; -- default rank for no title is ZERO
		rgNew[0] = { name = L["None"], id = -1, rank = iRank };
		gv.rgstrTitleNames[0] = L["None"];
	end

	-- build a table for reverse lookups, to go from title id to sorted order index
	for key, value in pairs(rgNew) do
		gv.rgTitleMap[value.id] = key;
	end

	-- now that the map is set, we can properly init gv.iTitleForRanking if
	-- we haven't already pointed it at a newly acquired title
	if (gv.iTitleForRanking == 0) then self:GetCurrentTitle(); end

	-- now that we've read everything in, blow away old saved ranking with new list
	self.db.profile.rgTitleInfo = rgNew;

	return true;
end


function Pokedex:GetTitleRankFromSavedData(iStart, idTitle)
	local cLimit = #self.db.profile.rgTitleInfo
	local index

	if (iStart <= cLimit) then
		for index = iStart, cLimit, 1 do
			if (self.db.profile.rgTitleInfo[index].id == idTitle) then
				return true, self.db.profile.rgTitleInfo[index].rank
			end
		end

		for index = 1, iStart - 1, 1 do
			if (self.db.profile.rgTitleInfo[index].id == idTitle) then
				return true, self.db.profile.rgTitleInfo[index].rank
			end
		end
	end

	return false, gc.iDefaultRank
end


--=========================================================================--
--=========================================================================--
--
-- AUTO-DISMOUNT FUNCTIONS
--
--=========================================================================--
--=========================================================================--

function Pokedex:CVAR_UPDATE(_, glstr, ...)
	if (DC.DISMOUNT >= DL.EXCEP) then self:Printf("CVAR_UPDATE EVENT:  %s %s", tostring(glstr), self:StrFromVarArg(nil, ...)); end
	if (glstr == "AUTO_DISMOUNT_FLYING_TEXT") then
		self:UpdateDismountSettings();
	end
end


function Pokedex:SetManageAutoDismount(info, value)
	if (gv.iAutoDismountFlying == 1 and value) then
		if (DC.DISMOUNT >= DL.EXCEP) then self:Print("Safe Dismount enabled, disabling Auto Dismount in Flight accordingly."); end
		gv.iAutoDismountFlying = 0;
		SetCVar("autoDismountFlying", 0); --, "AUTO_DISMOUNT_FLYING_TEXT")
	end

	Pokedex.db.profile.fManageAutoDismount = value;
end


function Pokedex:UpdateDismountSettings()
	if (GetCVarBool("autoDismountFlying")) then 
		-- even though we may be reseting it back to 0, our internal state should match reality
		gv.iAutoDismountFlying = 1;

		if (self.db.profile.fManageAutoDismount) then
			if (DC.DISMOUNT >= DL.EXCEP) then self:Print("Safe Dismount is reverting someones change to the Auto Dismount in Flight setting."); end
			SetCVar("autoDismountFlying", 0); --, "AUTO_DISMOUNT_FLYING_TEXT")
		end
	else
		gv.iAutoDismountFlying = 0;
	end
	LibStub("AceConfigRegistry-3.0"):NotifyChange("Pokedex");
end


function Pokedex:MainTooltipShow(...)
	-- if we have gathering skills and the feature is turned on
	-- if we're not currently flying then we don't have to worry about monkeying with the setting at all
	if (self.db.profile.fManageAutoDismount and Pokedex.db.profile.fDismountForGathering and IsFlying()) then
		-- if dismount was turned on for a different condition, tooltip scraping will only cause errors
		if ((not self.db.profile.fDismountForCombat or not gv.fCanDismountForCombat) and
		    (not self.db.profile.fDismountForAttack or not gv.fCanDismountForAttack)) then

			local fGatherable = false;
			for i=1,GameTooltip:NumLines() do
				local mytext = _G["GameTooltipTextLeft" .. i];
				local text = mytext:GetText();
				fGatherable = (gv.fMiner     and strfind(text, L["Requires"]) and strfind(text, L["Mining"])) or
				              (gv.fHerbalist and strfind(text, L["Requires"]) and strfind(text, L["Herbalism"])) or
				              (gv.fSkinner   and strfind(text, L["Skinnable"]));
			end

			if (fGatherable and gv.iAutoDismountFlying == 0) then
				if (DC.DISMOUNT >= DL.BASIC) then self:Print("turning on dismount, mousing over gatherable"); end
				gv.iAutoDismountFlying = 1;
				SetCVar("autoDismountFlying", 1); --, "AUTO_DISMOUNT_FLYING_TEXT")
			elseif (fGatherable and gv.iAutoDismountFlying == 1) then
				if (DC.DISMOUNT >= DL.EXCEP) then self:Print("WEIRD - new show, but can already dismount"); end
			elseif (not fGatherable and gv.iAutoDismountFlying == 1) then
				if (DC.DISMOUNT >= DL.BASIC) then self:Print("ERROR - can dismount, but tooltip doesn't match"); end
				gv.iAutoDismountFlying = 0;
				SetCVar("autoDismountFlying", 0); --, "AUTO_DISMOUNT_FLYING_TEXT")
			end
		end
	end
end


function Pokedex:MainTooltipHide(...)
	-- we only care if gathering feature is turned on
	if (self.db.profile.fManageAutoDismount and Pokedex.db.profile.fDismountForGathering) then
		-- if dismount is already off, we don't have to do anything
		if (gv.iAutoDismountFlying == 1) then
			-- if dismount was turned on for a different condition, tooltip scraping will only cause errors
			if ((not self.db.profile.fDismountForCombat or not gv.fCanDismountForCombat) and
			    (not self.db.profile.fDismountForAttack or not gv.fCanDismountForAttack)) then
				if (DC.DISMOUNT >= DL.BASIC) then self:Print("dismount turned off, not mousing over gatherable"); end
				gv.iAutoDismountFlying = 0;
				SetCVar("autoDismountFlying", 0); --, "AUTO_DISMOUNT_FLYING_TEXT")
			end
		end
	end	
end


function Pokedex:PLAYER_TARGET_CHANGED()
	-- if feature is turned on
	if (self.db.profile.fManageAutoDismount and self.db.profile.fDismountForAttack) then
		-- if we had an attackable target
		if (gv.fCanDismountForAttack) then
			-- then only need to change if we lost one
			if (not UnitCanAttack("player", "target") or UnitIsDead("target")) then
				gv.fCanDismountForAttack = false;

				-- if in combat, don't clear based on target
				if (not self.db.profile.fDismountForCombat or not gv.fCanDismountForCombat) then
					if (DC.DISMOUNT >= DL.BASIC) then self:Print("dismount turned off, target not attackable"); end
					gv.iAutoDismountFlying = 0;
					SetCVar("autoDismountFlying", 0); --, "AUTO_DISMOUNT_FLYING_TEXT")
				end
			end
		-- if not already tracking, then don't start unless we're mounted - hopeful optimization for instances
		elseif (IsMounted()) then
			-- if we have an attackable target
			if (UnitCanAttack("player", "target") and not UnitIsDead("target")) then
				gv.fCanDismountForAttack = true;
				
				-- don't need to set if already done for being in combat
				if (not self.db.profile.fDismountForCombat or not gv.fCanDismountForCombat) then
					if (DC.DISMOUNT >= DL.BASIC) then self:Print("turning on dismount, attackable target"); end
					gv.iAutoDismountFlying = 1;
					SetCVar("autoDismountFlying", 1); --, "AUTO_DISMOUNT_FLYING_TEXT")
				end
			end
		end
	end
end


function Pokedex:PLAYER_REGEN_DISABLED()
	-- if feature is turned on
	-- if we're not currently mounted then we don't have to worry about monkeying with the setting at all
	if (self.db.profile.fManageAutoDismount and self.db.profile.fDismountForCombat and IsMounted()) then
		gv.fCanDismountForCombat = true;

		if (gv.iAutoDismountFlying == 0) then
			if (DC.DISMOUNT >= DL.BASIC) then self:Print("turning on dismount, in combat"); end
			gv.iAutoDismountFlying = 1;
			SetCVar("autoDismountFlying", 1); --, "AUTO_DISMOUNT_FLYING_TEXT")
		end
	end
end


function Pokedex:PLAYER_REGEN_ENABLED()
	-- if feature is turned on
	if (self.db.profile.fManageAutoDismount and self.db.profile.fDismountForCombat) then
		-- out of combat, so this will no longer be the reason why we can dismount
		gv.fCanDismountForCombat = false;

		-- if can dismount is already 0, then there's nothing left to change
		if (gv.iAutoDismountFlying == 1) then
			-- if dismount can occur for attackable target, don't clear for out of combat
			if (not self.db.profile.fDismountForAttack or not gv.fCanDismountForAttack) then
				if (DC.DISMOUNT >= DL.BASIC) then self:Print("dismount turned off, out of combat"); end
				gv.iAutoDismountFlying = 0;
				SetCVar("autoDismountFlying", 0); --, "AUTO_DISMOUNT_FLYING_TEXT")
			end
		end
	end	
end
